/*
 * Tapioca daemon
 * Copyright (C) 2005 INdT 
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>
#include <glib-object.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include <sys/types.h>

#include "tapiocad_config.h"
#include "tapiocad_connection_manager.h"

/* variables */
static GMainLoop *mainloop;
static int daemon_pipe_fd[2];
static gboolean child_died = FALSE;
static TapiocadConnectionManager *tcm;

static void tapiocad_exit                  (void);
static void name_owner_changed             (DBusGProxy *proxy,
                                            const char *name,
                                            const char *prev_owner,
                                            const char *new_owner,
                                            gpointer user_data);
static int  tapiocad_parent_wait_for_child (int child_fd, 
                                            pid_t child_pid, 
                                            int timeout);
static void tapiocad_handle_signal         (int value);
static void tapiocad_handle_sigchld        (int value);

static void
usage (void)
{
    g_print ("usage: tapioca [--daemon] [--verbose] [--help]\n");
    g_print ("\n"
             "\t--daemon      Become a daemon\n"
             "\t--verbose     Turn on verbose mode\n"
             "\t--help        Show this information and exit\n");
}

int main (int argc, char **argv)
{
    gboolean daemon = FALSE;
    gboolean verbose = FALSE;
    DBusGConnection *bus;
    DBusGProxy *proxy;
    GError *error = NULL;
    guint32 request_name_ret;

    char option;
    struct option long_opts[] = {
        { "daemon",  0, NULL, 'd' },
        { "verbose", 0, NULL, 'v' },
        { "help",    0, NULL, 'h' },
        { NULL,      0, NULL, 0 }
    };

    g_type_init ();
    mainloop = g_main_loop_new (NULL, FALSE);
    
    // parse args
    while ((option = getopt_long (argc, argv, "", long_opts, NULL)) != -1) {
        switch (option) {
            case 'd':
                daemon = TRUE;
                break;
            case 'v':
                verbose = TRUE;
                break;
            case 'h':
                usage ();
                return 1;
            default:
                break;
        }
    }

    if (daemon) {
        int child_pid;
        char buf[1] = { 0 };

        if (pipe (daemon_pipe_fd) != 0) {
            g_print ("%s (%d): Could not setup pipe: '%s'\n", 
                    __FUNCTION__,
                    __LINE__,
                    strerror(errno));
            exit (1);
        }

        child_pid = fork ();
        switch (child_pid) {
            case -1: 
                g_print ("%s (%d): Could not fork process: '%s'\n", 
                        __FUNCTION__, 
                        __LINE__,
                        strerror (errno));
                exit (1);

            case 0: /* child process */ 
                break;

            default: /* parent process */
                exit (tapiocad_parent_wait_for_child (
                            daemon_pipe_fd[0], 
                            child_pid, 
                            5));
                break;
        };

        setsid ();
        write (daemon_pipe_fd[1], buf, sizeof (buf));
        close (daemon_pipe_fd[0]);
        close (daemon_pipe_fd[1]);
    }
       
    /* Obtain a connection to the session bus */
    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
    if (bus == NULL) {
        g_warning ("Failed to open connection to bus: %s", error->message);
        g_error_free (error);
        return 1;
    }

    if (!tapiocad_config_init ()) {
        g_warning ("No config found. Aborting!");
        return 1;
    }
 
    proxy = dbus_g_proxy_new_for_name (bus,
            DBUS_SERVICE_DBUS,
            DBUS_PATH_DBUS,
            DBUS_INTERFACE_DBUS);

    if (!org_freedesktop_DBus_request_name (proxy, "org.tapioca.Server",
                0, &request_name_ret, &error)) {
        g_warning ("Failed to get name: %s", error->message);
        return 1;
    }
   
    if (request_name_ret != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {
        g_warning ("Got result code %u from requesting name", request_name_ret);
        g_error_free (error);
        exit (1);
    }

    tcm = TAPIOCAD_CONNECTION_MANAGER (g_object_new (TAPIOCAD_CONNECTION_MANAGER_TYPE, NULL));

    dbus_g_proxy_add_signal (proxy, "NameOwnerChanged",
        G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INVALID);
    dbus_g_proxy_connect_signal (proxy, "NameOwnerChanged",
        G_CALLBACK (name_owner_changed), NULL, NULL);
    
    dbus_g_connection_register_g_object (bus,
            "/org/tapioca/ConnectionManager",
            G_OBJECT (tcm));
   
    /* Finally, setup unix signal handler for TERM */
    signal (SIGTERM, tapiocad_handle_signal);
    signal (SIGINT, tapiocad_handle_signal);
 
    g_main_loop_run (mainloop);

    return 0;    
}

/* internal */
static void
tapiocad_exit (void)
{
    g_object_unref (tcm);
    tapiocad_config_deinit ();
   
    g_main_loop_quit (mainloop);
}

static void
name_owner_changed (DBusGProxy *proxy,
                    const char *name,
                    const char *prev_owner,
                    const char *new_owner,
                    gpointer user_data)
{
    /* check if the service that exited was launched by tapiocad
     * and if so, remove it from services list.
     * Close tapiocad if no more service is loaded.
     */
    if (!strcmp (new_owner, "") && tapiocad_connection_manager_has_service (tcm, name)) {
        tapiocad_connection_manager_remove_service (tcm, name);
        if (tapiocad_connection_manager_get_service_count (tcm) == 0) {
            /* disconnect NameOwnerChanged signal so we don't receive it
             * when exiting.
             */
            dbus_g_proxy_disconnect_signal (proxy, "NameOwnerChanged", 
                    G_CALLBACK (name_owner_changed), NULL);
            tapiocad_exit ();
        }
    }
}

static int
tapiocad_parent_wait_for_child (int child_fd, 
                                pid_t child_pid, 
                                int timeout)
{
    fd_set rfds;
    fd_set efds;
    struct timeval tv;
    int retval;

    signal (SIGCHLD, tapiocad_handle_sigchld);

    FD_ZERO (&rfds);
    FD_SET (child_fd, &rfds);
    FD_ZERO (&efds);
    FD_SET (child_fd, &efds);
    
    tv.tv_sec = timeout;
    tv.tv_usec = 0;

    retval = select (child_fd + 1, &rfds, NULL, &efds, &tv);

    /* child has died */
    if (child_died) 
        return 1;

    /* everything runs fine */
    if (retval > 0)
        return 0;

    /* timeout; kill child */
    kill (child_pid, SIGTERM);    
    return 1;
}

static void
tapiocad_handle_signal (int value)
{
    tapiocad_exit ();
}

static void
tapiocad_handle_sigchld (int value)
{
    child_died = TRUE;
}
