/*
 * Tapioca Daemon
 * Copyright (c) 2005 INdT.
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "tapiocad_config.h"

#include <stdio.h>
#include <string.h>
#include <glib/gmem.h>
#include <glib/gmessages.h>
#include <glib/gstrfuncs.h>

struct _TapiocadConfigPrivate 
{
    gboolean initialized;
    gchar *current_group;
    GHashTable *hash_group;
    GHashTable *current_hash_entry;
};

static void                  tapiocad_config_load_all            (void);

static TapiocadConfig       *tapiocad_config_new                 (const gchar *file);
static void                  tapiocad_config_destroy             (TapiocadConfig *config);
static gboolean              tapiocad_config_parse_file          (TapiocadConfig *config);
static void                  tapiocad_config_set_group           (TapiocadConfig *config, 
                                                                  const gchar   *group);
static G_CONST_RETURN gchar *tapiocad_config_read_entry          (TapiocadConfig *config,
                                                                  const gchar   *entry,
                                                                  const gchar   *default_value);

static void                  tapiocad_config_cleanup             (TapiocadConfig *config);
static void                  tapiocad_config_hash_free_str_func  (gpointer data);
static void                  tapiocad_config_hash_free_hash_func (gpointer data);

/* TODO create function to handle search paths */
static guint  search_path_count = 1;
static gchar *search_paths[] = {
    CONFIGDIR
};

static GList *configs = NULL;

gboolean
tapiocad_config_init (void)
{
#if 1
    GList *l;
    TapiocadConfig *config;
    guint i, protocols_length;
#endif
    
    tapiocad_config_load_all ();

    if (g_list_length (configs) == 0)
        return FALSE;
   
#if 1
    g_print ("services: '%d'\n", 
            g_list_length (configs));

    for (l = configs; l != NULL; l = l->next) {
        config = (TapiocadConfig *) l->data;
        g_print ("\n"
        "\tservice_name: %s\n"
        "\tsupported protocols: ",
        config->service_name);
        protocols_length = g_strv_length (config->protocols);
        for (i = 0; i < protocols_length; ++i) 
            g_print ("%s ", config->protocols[i]);
        g_print ("\n");
    }
    g_print ("\n");
#endif

    return TRUE;
}

void
tapiocad_config_deinit (void)
{
    GList *l;
    TapiocadConfig *config;
    
    g_return_if_fail (configs != NULL);

    for (l = configs; l != NULL; l = l->next) {
        config = l->data;
        tapiocad_config_destroy (config);
        g_free (config);
    }

    g_list_free (configs);
    configs = NULL;
}

G_CONST_RETURN GList *
tapiocad_config_get_list (void)
{
    return configs;
}

TapiocadConfig *
tapiocad_config_get_from_protocol (const gchar *protocol)
{
    GList *l;
    TapiocadConfig *config;
    guint i, protocols_length;
    
    g_return_val_if_fail (configs != NULL, NULL);
    g_return_val_if_fail (protocol != NULL, NULL);

    for (l = configs; l != NULL; l = l->next) {
        config = (TapiocadConfig *) l->data;
        protocols_length = g_strv_length (config->protocols);
        for (i = 0; i < protocols_length; ++i) {
            if (g_strcasecmp (config->protocols[i], protocol) == 0) {
                return config;
            }
        }
    }
    
    // no config to handle this protocol
        
    return NULL;
}

static void
tapiocad_config_load_all (void)
{
    guint i;
    GDir *dir;
    const gchar *file;
    gchar *path;
    TapiocadConfig *config;
   
    for (i = 0; i < search_path_count; ++i) {
        dir = g_dir_open (search_paths[i], 0, NULL);
        if (dir == NULL)
            continue;
            
        while ((file = g_dir_read_name (dir)) != NULL) {
            if (g_str_has_suffix (file, ".ini")) {
                path = g_build_filename (search_paths[i], file, NULL);
                if ((config = tapiocad_config_new (path)))
                    configs = g_list_append (configs, config);            
                g_free (path);
            }
        }
        
        g_dir_close (dir);
    }
}

// TapiocadConfig object
TapiocadConfig *
tapiocad_config_new (const gchar *file)
{
    TapiocadConfig *config;
    const gchar *protocols;
    
    g_return_val_if_fail (file != NULL, NULL);
    
    config = g_new0 (TapiocadConfig, 1);
    config->file = g_strdup (file);
    
    config->priv = g_new0 (TapiocadConfigPrivate, 1);
    config->priv->initialized = FALSE;
    config->priv->current_group = NULL;
    config->priv->hash_group = NULL;
    config->priv->current_hash_entry = NULL;
    
    if (!tapiocad_config_parse_file (config)) {
        g_free (config);
        return NULL;
    }

    tapiocad_config_set_group (config, "Defaults");
    
    config->service_name = g_strdup (tapiocad_config_read_entry (config, "Service Name", NULL));
    if (!config->service_name)
        goto error;
    
    protocols = tapiocad_config_read_entry (config, "Protocols", NULL);
    if (!protocols) 
        goto error;
    config->protocols = g_strsplit (protocols, ",", -1);

    tapiocad_config_cleanup (config);
    
    return config;

error:
    g_debug ("invalid config file %s", config->file);
    tapiocad_config_cleanup (config);
    tapiocad_config_destroy (config);
    return NULL;
}

void 
tapiocad_config_destroy (TapiocadConfig *config)
{
    g_return_if_fail (config != NULL);
    
    g_free (config->file);
   
    if (config->service_name)
        g_free (config->service_name);
    if (config->protocols)
        g_strfreev (config->protocols);

    g_free (config->priv);
}

gboolean 
tapiocad_config_parse_file (TapiocadConfig *config)
{
    FILE *file;
    gchar line[256];
    gchar *group, *key, *value;
#ifdef DEBUG
    guint current_line = 0;
#endif  
  
    g_return_val_if_fail (config != NULL, FALSE);
    
    if (config->priv->initialized)
        tapiocad_config_cleanup (config);  
    
    file = fopen (config->file, "rb");
    if (!file) {
        // error - file not found or permission denied
        return FALSE;
    }
    
    config->priv->hash_group = g_hash_table_new_full (g_str_hash, 
            g_str_equal, 
            tapiocad_config_hash_free_str_func,
            tapiocad_config_hash_free_hash_func);
        
    while (!feof (file)) {
#ifdef DEBUG
        ++current_line;
#endif  

        fgets (line, 256, file);
        g_strchug (line);
        if (line[0] == '\0' || line[0] == '#')
            continue;

        g_strchomp (line);
        
        // group found
        if (line[0] == '[') {
            if (line[strlen (line) - 1] != ']') {
#ifdef DEBUG      
                g_print ("invalid group '%s' at line '%d'\n", line, current_line);
#endif
                
                tapiocad_config_cleanup (config);
                fclose (file);
                return FALSE;
            }
            
            group = g_new0 (gchar, strlen (line) - 1);
            g_strlcpy (group, line + 1, strlen(line) - 1);
     
            config->priv->current_hash_entry = g_hash_table_new_full (g_str_hash, 
                    g_str_equal, 
                    tapiocad_config_hash_free_str_func,
                    tapiocad_config_hash_free_str_func);

            g_hash_table_insert (config->priv->hash_group,
                    group, 
                    config->priv->current_hash_entry);
            continue;
        }
        
        value = strchr (line, '=');
        if (!value) { 
#ifdef DEBUG      
            g_print ("invalid entry '%s' at line '%d'\n", line, current_line);
#endif
          
            tapiocad_config_cleanup (config);
            fclose (file);
            return FALSE;
        }
          
        if (config->priv->current_hash_entry == NULL) {
            // no group
            config->priv->current_hash_entry = g_hash_table_new_full (g_str_hash, 
                    g_str_equal, 
                    tapiocad_config_hash_free_str_func,
                    tapiocad_config_hash_free_str_func);
            g_hash_table_insert (config->priv->hash_group,
                    g_strdup(""), 
                    config->priv->current_hash_entry);
        }
       
        key = g_new0 (gchar, value - line + 1);
        g_strlcpy (key, line, value - line + 1);
        
        // TODO test duplicate entry 

        g_hash_table_insert (config->priv->current_hash_entry, 
                key, 
                g_strdup (value + 1));
    }

    config->priv->current_hash_entry = NULL; 
    config->priv->initialized = TRUE;
    
    fclose (file);
    return TRUE;
}

void 
tapiocad_config_set_group (TapiocadConfig *config, 
                           const gchar   *group)
{
    g_return_if_fail (config != NULL);
    
    if (config->priv->current_group)
        g_free (config->priv->current_group);
    
    if (group)
        config->priv->current_group = g_strdup (group);
    else 
        config->priv->current_group = NULL;
}

G_CONST_RETURN gchar *
tapiocad_config_read_entry (TapiocadConfig *config,
                            const gchar    *entry,
                            const gchar    *default_value)
{
    gchar *group;
    gpointer *key, *value;
    
    g_return_val_if_fail (config != NULL, NULL);
    g_return_val_if_fail (config->priv->initialized != FALSE, NULL);
    g_return_val_if_fail (entry != NULL, NULL);
    
    if (!config->priv->current_hash_entry) { // use default group
        if (config->priv->current_group)
            group = config->priv->current_group;
        else
            group = "";
          
        if (!g_hash_table_lookup_extended (config->priv->hash_group, group, (gpointer) &key, (gpointer) &value)) {
#ifdef DEBUG
            g_print ("%s (%d): invalid group '%s'. use tapiocad_config_set_group\n",
                    __FUNCTION__, __LINE__, group);
#endif
            return NULL;
        }
        config->priv->current_hash_entry = (GHashTable *) value;
    }
    
    if (g_hash_table_lookup_extended (config->priv->current_hash_entry, entry, (gpointer) &key, (gpointer) &value)) {
        return (gchar *) value;
    }

    return default_value;
}

static void 
tapiocad_config_cleanup (TapiocadConfig *config)
{
    if (!config->priv->initialized)
        return;
    
    config->priv->initialized = FALSE;
    if (config->priv->current_group) {
        g_free (config->priv->current_group);
        config->priv->current_group = NULL;
    }
    
    g_hash_table_destroy (config->priv->hash_group);
    config->priv->hash_group = NULL;
    config->priv->current_hash_entry = NULL;
}

static void 
tapiocad_config_hash_free_str_func (gpointer data)
{ 
    g_free (data);
}

static void 
tapiocad_config_hash_free_hash_func (gpointer data)
{
    g_hash_table_destroy ((GHashTable *) data);
}
