/*
 * Tapioca daemon
 * Copyright (C) 2005 INdT 
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapiocad_connection_manager.h"
#include "tapiocad_connection_manager_glue.h"

#include "tapiocad_config.h"

#include "tapioca_connection_manager_bindings.h"

#include <tapioca/core/tapioca_errors.h>

struct _TapiocadConnectionManagerPrivate {
    GHashTable *services;
    gboolean disposed;
};

static void tapiocad_connection_manager_dispose (GObject *object);

G_DEFINE_TYPE(TapiocadConnectionManager, tapiocad_connection_manager, G_TYPE_OBJECT)
   
static void
tapiocad_connection_manager_class_init (TapiocadConnectionManagerClass *klass)
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapiocad_connection_manager_parent_class = g_type_class_peek_parent (klass);
    
    g_type_class_add_private (klass, sizeof (TapiocadConnectionManagerPrivate));
    
    gobject_class->dispose = tapiocad_connection_manager_dispose;
    
    dbus_g_object_type_install_info (
            G_TYPE_FROM_CLASS (klass), 
            &dbus_glib_tapiocad_connection_manager_object_info);
}

static void
tapiocad_connection_manager_init (TapiocadConnectionManager *tcm)
{
    tcm->priv = TAPIOCAD_CONNECTION_MANAGER_GET_PRIVATE (tcm);
    tcm->priv->services = NULL;
    tcm->priv->disposed = FALSE;
}

static void 
tapiocad_connection_manager_dispose (GObject *object)
{
    TapiocadConnectionManager *tcm = TAPIOCAD_CONNECTION_MANAGER (object);

    if (tcm->priv->disposed) {
        /* If dispose did already run, return. */
        return;
    }
    
    /* Make sure dispose does not run twice. */
    tcm->priv->disposed = TRUE;

    if (tcm->priv->services)
        g_hash_table_destroy (tcm->priv->services);
    
    G_OBJECT_CLASS (tapiocad_connection_manager_parent_class)->dispose (object);
}

gboolean 
tapiocad_connection_manager_has_service (TapiocadConnectionManager *tcm,
                                         const gchar *service)
{
    if (!tcm->priv->services)
        return FALSE;
    return (g_hash_table_lookup (tcm->priv->services, (gpointer) service) != NULL); 
}

void
tapiocad_connection_manager_remove_service (TapiocadConnectionManager *tcm,
                                            const gchar *service)
{
    g_hash_table_remove (tcm->priv->services, (gpointer) service);
}

guint 
tapiocad_connection_manager_get_service_count (TapiocadConnectionManager *tcm)
{
    return g_hash_table_size (tcm->priv->services);
}

gboolean
tapiocad_connection_manager_get_supported_protocols (TapiocadConnectionManager *tcm, 
                                                     gchar ***protocols,
                                                     GError **error)
{
    const GList *configs, *configs_walk;
    TapiocadConfig *config;
    guint length = 0, count = 0, i;
    
    configs = tapiocad_config_get_list ();
    for (configs_walk = configs; configs_walk != NULL; configs_walk = g_list_next (configs_walk)) {
        config = configs_walk->data;
        length += g_strv_length (config->protocols);
    }

    *protocols = g_new (gchar *, length + 1);
    (*protocols)[length] = NULL;
    for (configs_walk = configs; configs_walk != NULL; configs_walk = g_list_next (configs_walk)) {
        config = configs_walk->data;
        length = g_strv_length (config->protocols);
        for (i = 0; i < length; ++i)
            (*protocols)[count++] = g_strdup (config->protocols[i]);
    }
        
    return TRUE;
}

gboolean
tapiocad_connection_manager_create_connection (TapiocadConnectionManager *tcm, 
                                               const gchar *protocol,
                                               gchar **service_name,
                                               gchar **object_path,
                                               GError **error)
{
    TapiocadConfig *config;
    DBusGConnection *bus;
    GError *bus_error = NULL;
    DBusGProxy *proxy;

    g_debug ("creating connection for protocol %s ...", protocol);
    
    config = tapiocad_config_get_from_protocol (protocol);
    if (!config) {
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_SUPPORTED, "Protocol not supported");  
        return FALSE;
    }        

    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &bus_error);
    if (!bus) {
        g_warning ("unable to open connection to bus: %s", bus_error->message);
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_ENOUGH_MEMORY, "Not enough memory");  
        return FALSE;
    }

    /* activate config service and create a new connection */
    proxy = dbus_g_proxy_new_for_name (bus,
            config->service_name,
            "/org/tapioca/ConnectionManager",
            "org.tapioca.ConnectionManager");
    
    if (!tcm->priv->services)
        tcm->priv->services = g_hash_table_new_full (g_str_hash, g_str_equal, 
                NULL, g_object_unref);
    g_hash_table_insert (tcm->priv->services, (gpointer) config->service_name, (gpointer) proxy);
    
    if (!org_tapioca_ConnectionManager_create_connection (proxy, object_path, error) ||
        !(*object_path)) {
        g_hash_table_remove (tcm->priv->services, (gpointer) config->service_name);
        
        if (*error == NULL)
            g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, 
                    "Cannot create connection object");
            
        g_warning ("error activating service %s - %s\n", config->service_name, (*error)->message);            

        return FALSE;
    }
       
    *service_name = g_strdup (config->service_name);
    g_debug ("connection created: service_name=%s object_path=%s", *service_name, *object_path);

    return TRUE;
}
