/*
 * Tapioca library
 * Copyright (C) 2005 INdT 
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapioca_audio_session.h"

#include "tapioca_session_private.h"
#include "tapioca_audio_session_private.h"
#include "tapioca_audio_session_bindings.h"
#include "tapioca_marshal.h"

#include <dbus/dbus-glib-bindings.h>

/* Signals */
enum {
    REMOTE_CAPABILITIES_UPDATED,
    REMOTE_CANDIDATE_ADDED,
    LAST_SIGNAL
};

/* Private members */
struct _TapiocaAudioSessionPrivate {
    DBusGProxy *proxy;
    gboolean disposed;
};

static void tapioca_audio_session_dispose     (GObject *object);
static void tapioca_audio_session_finalize    (GObject *object);

static void proxy_remote_capabilities_updated (DBusGProxy *proxy, 
                                               const gchar *remote_capabilities, 
                                               gpointer data);
static void proxy_remote_candidate_added      (DBusGProxy *proxy,
                                               const gchar *name,
                                               const gchar *protocol,
                                               const gchar *ip,
                                               guint port,
                                               gdouble preference,
                                               const gchar *username,
                                               const gchar *password,
                                               const gchar *type,
                                               guint generation,
                                               gpointer data);

G_DEFINE_TYPE(TapiocaAudioSession, tapioca_audio_session, TAPIOCA_TYPE_SESSION)

static guint tapioca_audio_session_signals[LAST_SIGNAL] = { 0 };

static void 
tapioca_audio_session_class_init (TapiocaAudioSessionClass *klass) 
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_audio_session_parent_class = g_type_class_peek_parent (klass);
    
    g_type_class_add_private (klass, sizeof (TapiocaAudioSessionPrivate));

    gobject_class->dispose = tapioca_audio_session_dispose;
    gobject_class->finalize = tapioca_audio_session_finalize;
    
    /* Create signals */
    tapioca_audio_session_signals[REMOTE_CAPABILITIES_UPDATED] = g_signal_new ("remote-capabilities-updated",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__STRING,
        G_TYPE_NONE,
        1,
        G_TYPE_STRING);
    tapioca_audio_session_signals[REMOTE_CANDIDATE_ADDED] = g_signal_new ("remote-candidate-added",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        tapioca_marshal_VOID__STRING_STRING_STRING_UINT_DOUBLE_STRING_STRING_STRING_UINT,
        G_TYPE_NONE,
        9,
        G_TYPE_STRING,
        G_TYPE_STRING,
        G_TYPE_STRING,
        G_TYPE_UINT,
        G_TYPE_DOUBLE,
        G_TYPE_STRING,
        G_TYPE_STRING,
        G_TYPE_STRING,
        G_TYPE_UINT);
}

static void 
tapioca_audio_session_init (TapiocaAudioSession *tas)
{
    tas->priv = TAPIOCA_AUDIO_SESSION_GET_PRIVATE (tas);
    tas->priv->proxy = NULL;
    tas->priv->disposed = FALSE;
}

static void 
tapioca_audio_session_dispose (GObject *object) 
{
    TapiocaAudioSession *tas = TAPIOCA_AUDIO_SESSION (object);
    
    if (tas->priv->disposed) {
        /* If dispose did already run, return. */
        return;
    }
  
    /* Make sure dispose does not run twice. */
    tas->priv->disposed = TRUE;
  
    if (tas->priv->proxy) {
        dbus_g_proxy_disconnect_signal (tas->priv->proxy, "RemoteCapabilitiesUpdated",
                G_CALLBACK (proxy_remote_capabilities_updated), NULL);
        dbus_g_proxy_disconnect_signal (tas->priv->proxy, "RemoteCandidateAdded",
                G_CALLBACK (proxy_remote_candidate_added), NULL);

        g_object_unref (tas->priv->proxy);
    }
    
    G_OBJECT_CLASS (tapioca_audio_session_parent_class)->dispose (object);
}

static void 
tapioca_audio_session_finalize (GObject *object) 
{
    g_signal_handlers_destroy (object);
    
    G_OBJECT_CLASS (tapioca_audio_session_parent_class)->finalize (object);
}

TapiocaAudioSession *
tapioca_audio_session_new (DBusGConnection *bus,
                           const gchar *service_name, 
                           const gchar *obj_path) 
{
    TapiocaAudioSession *tas;
    DBusGProxy *proxy;
   
    proxy = dbus_g_proxy_new_for_name (bus,
            service_name,
            obj_path,
            "org.tapioca.Session.Audio");
    if (!proxy) {
        g_warning ("failed to create proxy for service %s with object path %s on interface org.tapioca.Session.Audio", 
                service_name, obj_path);
        return NULL;
    }

    tas = TAPIOCA_AUDIO_SESSION (g_object_new (tapioca_audio_session_get_type (), NULL));
   
    if (!tapioca_session_setup (TAPIOCA_SESSION (tas), bus, service_name, obj_path)) {
        g_object_unref (tas);
        return NULL;
    }
 
    tas->priv->proxy = proxy;
 
    dbus_g_proxy_add_signal (tas->priv->proxy, "RemoteCapabilitiesUpdated", 
            G_TYPE_STRING, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (tas->priv->proxy, "RemoteCandidateAdded", 
            G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_DOUBLE, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_INVALID);
    dbus_g_proxy_connect_signal (tas->priv->proxy, "RemoteCapabilitiesUpdated",
            G_CALLBACK (proxy_remote_capabilities_updated),
            tas, NULL);
    dbus_g_proxy_connect_signal (tas->priv->proxy, "RemoteCandidateAdded",
            G_CALLBACK (proxy_remote_candidate_added),
            tas, NULL);
   
    return tas;
}

/**
 * Accept an incoming audio session.
 *
 * @param tas #TapiocaAudioSession instance.
 * @return Returns FALSE if an error occurs otherwise TRUE. 
 */
gboolean 
tapioca_audio_session_accept (TapiocaAudioSession *tas)
{
    GError *error = NULL;

    g_return_val_if_fail (tas != NULL, FALSE);
    // TODO check session state
    
    if (!(org_tapioca_Session_Audio_accept (tas->priv->proxy, &error))) {
        g_warning ("failed to accept a session");
        return FALSE;
    }
   
    return TRUE; 
}

/**
 * Reject an incoming audio session.
 *
 * @param tas #TapiocaAudioSession instance.
 * @return Returns FALSE if an error occurs otherwise TRUE.
 */
gboolean
tapioca_audio_session_reject (TapiocaAudioSession *tas)
{
    GError *error = NULL;
    
    g_return_val_if_fail (tas != NULL, FALSE);
    // TODO check session state
    
    if (!(org_tapioca_Session_Audio_reject (tas->priv->proxy, &error))) {
        g_warning ("failed to reject a session");
        return FALSE;
    }
   
    return TRUE; 
}

/**
 * Set local capabilities.
 *
 * @param tas #TapiocaAudioSession instance.
 * @param local_capabilities The local capabilities described as SDP.
 * @return Returns FALSE if an error occurs otherwise TRUE.
 */
gboolean
tapioca_audio_session_set_local_capabilities (TapiocaAudioSession *tas,
                                              const gchar *local_capabilities)
{
    GError *error = NULL;
    
    g_return_val_if_fail (tas != NULL, FALSE);
    g_return_val_if_fail (local_capabilities != NULL, FALSE);
    
    if (!(org_tapioca_Session_Audio_set_local_capabilities (
                    tas->priv->proxy, 
                    local_capabilities, &error))) {
        g_warning ("failed setting local capabilities");
        return FALSE;
    }
 
    return TRUE; 
}

/**
 * Get session capabilities.
 *
 * @param tas #TapiocaAudioSession instance.
 * @param local_capabilities The local capabilities.
 * @param remote_capabilities The remote capabilities.
 * @return Returns FALSE if an error occurs otherwise TRUE.
 */
gboolean
tapioca_audio_session_get_capabilities (TapiocaAudioSession *tas,
                                        gchar **local_capabilities,
                                        gchar **remote_capabilities)
{
    GError *error = NULL;
    
    g_return_val_if_fail (tas != NULL, FALSE);
    
    if (!(org_tapioca_Session_Audio_get_capabilities (
                    tas->priv->proxy, 
                    local_capabilities, 
                    remote_capabilities, 
                    &error))) {
        g_warning ("failed getting capabilities");
        return FALSE;
    }
    
    return TRUE; 
}

gboolean
tapioca_audio_session_add_local_candidate (TapiocaAudioSession *tas, 
                                           const gchar *name,
                                           const gchar *protocol,
                                           const gchar *ip,
                                           guint port,
                                           gdouble preference,
                                           const gchar *username,
                                           const gchar *password,
                                           const gchar *type,
                                           guint generation)
{
    GError *error = NULL;
    
    g_return_val_if_fail (tas != NULL, FALSE);
 
    if (!org_tapioca_Session_Audio_add_local_candidate (
                tas->priv->proxy, 
                name,
                protocol,
                ip,
                port,
                preference,
                username,
                password,
                type,
                generation,
                &error)) {
        g_warning ("failed to add local candidate");
        return FALSE;
    }
    
    return TRUE;
}

static void
proxy_remote_capabilities_updated (DBusGProxy *proxy,
                                   const gchar *remote_capabilities,
                                   gpointer data)
{
    g_signal_emit (TAPIOCA_AUDIO_SESSION (data), 
            tapioca_audio_session_signals[REMOTE_CAPABILITIES_UPDATED], 
            0, remote_capabilities);
}

static void
proxy_remote_candidate_added (DBusGProxy *proxy,
                              const gchar *name,
                              const gchar *protocol,
                              const gchar *ip,
                              guint port,
                              gdouble preference,
                              const gchar *username,
                              const gchar *password,
                              const gchar *type,
                              guint generation,
                              gpointer data)
{
    g_signal_emit (TAPIOCA_AUDIO_SESSION (data), 
            tapioca_audio_session_signals[REMOTE_CANDIDATE_ADDED],
            0,
            name,
            protocol,
            ip,
            port,
            preference,
            username,
            password,
            type,
            generation);
}
