/*
 * Tapioca library
 * Copyright (C) 2005 INdT 
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 * @author Renato Araujo Oliveira Filho <renato.filho@indt.org.br>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapioca_connection.h"

#include "tapioca_connection_bindings.h"
#include "tapioca_contact_list_private.h"
#include "tapioca_presence_private.h"
#include "tapioca_audio_session_private.h"
#include "tapioca_text_session_private.h"
#include "tapioca_marshal.h"

#include <dbus/dbus-glib-bindings.h>

/* Signals */
enum {
    STATE_CHANGED,
    ERROR,
    INCOMING_SESSION,
    LAST_SIGNAL
};

struct _TapiocaConnectionPrivate {
    DBusGConnection *bus;
    DBusGProxy *proxy;
    gchar *service_name;
    gchar *id;
    TapiocaConnectionState state;
    TapiocaPresence *presence;
    TapiocaContactList *contact_list;
    gboolean disposed;
};

static void     tapioca_connection_dispose       (GObject *object);
static void     tapioca_connection_finalize      (GObject *object);

static void     destroy_connection               (TapiocaConnection *tc);
static void     proxy_state_changed_signal       (DBusGProxy *proxy,
                                                  guint state,
                                                  gpointer data);
static void     proxy_incoming_session_signal    (DBusGProxy *proxy,
                                                  const gchar *from,
                                                  guint type,
                                                  const gchar *obj_path,
                                                  gpointer data);
static void     proxy_error_signal               (DBusGProxy *proxy,
                                                  guint code,
                                                  gpointer data);

G_DEFINE_TYPE(TapiocaConnection, tapioca_connection, G_TYPE_OBJECT)

static guint tapioca_connection_signals[LAST_SIGNAL] = { 0 };

static void 
tapioca_connection_class_init (TapiocaConnectionClass *klass) 
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_connection_parent_class = g_type_class_peek_parent (klass);

    g_type_class_add_private (klass, sizeof (TapiocaConnectionPrivate));
    
    gobject_class->dispose = tapioca_connection_dispose;
    gobject_class->finalize = tapioca_connection_finalize;

    /* Let's create our signals */
    tapioca_connection_signals[STATE_CHANGED] = g_signal_new ("state-changed",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__UINT,
        G_TYPE_NONE,
        1,
        G_TYPE_UINT);
    tapioca_connection_signals[ERROR] = g_signal_new ("error",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__UINT,
        G_TYPE_NONE,
        1,
        G_TYPE_UINT);
    tapioca_connection_signals[INCOMING_SESSION] = g_signal_new ("incoming-session",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        tapioca_marshal_VOID__STRING_UINT_OBJECT,
        G_TYPE_NONE,
        3,
        G_TYPE_STRING,
        G_TYPE_UINT,
        G_TYPE_OBJECT);
}

static void 
tapioca_connection_init (TapiocaConnection *tc) 
{
    tc->priv = TAPIOCA_CONNECTION_GET_PRIVATE (tc);
    tc->priv->bus = NULL;
    tc->priv->proxy = NULL;
    tc->priv->service_name = NULL;
    tc->priv->id = NULL;
    tc->priv->state = TAPIOCA_CONNECTION_STATE_DISCONNECTED;
    tc->priv->presence = NULL;
    tc->priv->contact_list = NULL;
    tc->priv->disposed = FALSE;
}

static void 
tapioca_connection_dispose (GObject *object) 
{
    TapiocaConnection *tc = TAPIOCA_CONNECTION (object);

    if (tc->priv->disposed) 
       /* If dispose did already run, return. */
       return;

    /* Make sure dispose does not run twice. */
    tc->priv->disposed = TRUE;

    if (tc->priv->proxy) {
        dbus_g_proxy_disconnect_signal (tc->priv->proxy, "StateChanged",
                G_CALLBACK (proxy_state_changed_signal), tc);
        dbus_g_proxy_disconnect_signal (tc->priv->proxy, "IncomingSession",
                G_CALLBACK (proxy_incoming_session_signal), tc);
        dbus_g_proxy_disconnect_signal (tc->priv->proxy, "Error",
                G_CALLBACK (proxy_error_signal), tc);
   
        if (tc->priv->state != TAPIOCA_CONNECTION_STATE_DISCONNECTED) 
            destroy_connection (tc);

        g_object_unref (tc->priv->proxy);
    }

    g_free (tc->priv->service_name);
    if (tc->priv->id)
        g_free (tc->priv->id);
    if (tc->priv->presence)
        g_object_unref (tc->priv->presence);
    if (tc->priv->contact_list)
        g_object_unref (tc->priv->contact_list);

    G_OBJECT_CLASS (tapioca_connection_parent_class)->dispose (object);
}
        
static void 
tapioca_connection_finalize (GObject *object) 
{
    g_signal_handlers_destroy (object);

    G_OBJECT_CLASS (tapioca_connection_parent_class)->finalize (object);
}

TapiocaConnection *
tapioca_connection_new (DBusGConnection *bus,
                        const gchar *service_name,
                        const gchar *obj_path) 
{
    TapiocaConnection *tc;
    DBusGProxy *proxy;
    gchar *aux_obj_path;
    GError *error = NULL;
    
    proxy = dbus_g_proxy_new_for_name (bus,
            service_name,
            obj_path,
            "org.tapioca.Connection");
    if (!proxy) {
        g_warning ("failed to create proxy for service %s with object path %s on interface org.tapioca.Connection", 
                service_name, obj_path);
        return NULL;
    }

    tc = TAPIOCA_CONNECTION (g_object_new (tapioca_connection_get_type (), NULL));
    tc->priv->bus = bus;
    tc->priv->proxy = proxy;
    tc->priv->service_name = g_strdup (service_name);
 
    /* Add all DBus signals */
    dbus_g_proxy_add_signal (tc->priv->proxy, "StateChanged", 
            G_TYPE_UINT, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (tc->priv->proxy, "Error", 
            G_TYPE_UINT, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (tc->priv->proxy, "IncomingSession", 
            G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_INVALID);

    /* Connect all DBus signals */
    dbus_g_proxy_connect_signal (tc->priv->proxy, "StateChanged",
            G_CALLBACK (proxy_state_changed_signal),
            tc, NULL);
    dbus_g_proxy_connect_signal (tc->priv->proxy, "Error",
            G_CALLBACK (proxy_error_signal),
            tc, NULL);
    dbus_g_proxy_connect_signal (tc->priv->proxy, "IncomingSession",
            G_CALLBACK (proxy_incoming_session_signal),
            tc, NULL);

    /* create a new presence instance */
    if ((org_tapioca_Connection_get_presence (tc->priv->proxy,
                    &aux_obj_path,
                    &error))) {
        tc->priv->presence = tapioca_presence_new (tc->priv->bus, tc->priv->service_name, aux_obj_path);
    }

    if (aux_obj_path) {
        g_free (aux_obj_path);
        aux_obj_path = NULL;
    }
 
    /* create a new contact list instance */
    if ((org_tapioca_Connection_get_contact_list (tc->priv->proxy,
                    &aux_obj_path,
                    &error))) {
        tc->priv->contact_list = tapioca_contact_list_new (tc->priv->bus, tc->priv->service_name, aux_obj_path);
    }

    if (aux_obj_path) {
        g_free (aux_obj_path);
        aux_obj_path = NULL;
    }
   
    return tc;
}

/** 
 * Start connection.
 *
 * @param tc #TapiocaConnection instance.
 * @param id connection id.
 * @param table necessary parameters to make a connection.
 * @return FALSE if an error occurs otherwise TRUE
 */
gboolean
tapioca_connection_connect (TapiocaConnection *tc,
                            const gchar *id,
                            const GHashTable *table)
{
    GError *error = NULL;

    g_return_val_if_fail (tc != NULL, FALSE);
    g_return_val_if_fail (id != NULL, FALSE);
    g_return_val_if_fail (table != NULL, FALSE);
    g_return_val_if_fail (tc->priv->state == TAPIOCA_CONNECTION_STATE_DISCONNECTED, FALSE);

    /* try to connect */
    if (!(org_tapioca_Connection_connect (tc->priv->proxy, 
                    id, table, &error))) {
        g_warning ("error on connect");
        // TODO treat error as apropriate
        return FALSE;
    }

    tc->priv->id = g_strdup (id);

    return TRUE;
}

/**
 * Finish connection.
 *
 * @param tc #TapiocaConnection instance.
 */
void
tapioca_connection_disconnect (TapiocaConnection *tc)
{
    GError *error = NULL;

    g_return_if_fail (tc != NULL);
    g_return_if_fail (tc->priv->state != TAPIOCA_CONNECTION_STATE_DISCONNECTED);
    
    g_free (tc->priv->id);
    tc->priv->id = NULL;
    
    org_tapioca_Connection_disconnect (tc->priv->proxy, &error);
}

/**
 * Gets #TapiocaConnection state.
 *
 * @param tc #TapiocaConnection instance
 * @return #TapiocaConnectionState state of instance
 */
TapiocaConnectionState
tapioca_connection_get_state (TapiocaConnection *tc)
{
    g_return_val_if_fail (tc != NULL, -1);
    return tc->priv->state;
}

/**
 * Create a new #TapiocaSession.
 *
 * @param tc #TapiocaConnection instance.
 * @param session_type #TapiocaSessionType of thession session that will be created.
 * @return #TapiocaSession instance or NULL if fail.
 */
TapiocaSession *
tapioca_connection_create_session (TapiocaConnection *tc, 
                                   TapiocaSessionType session_type)
{
    TapiocaSession *ts = NULL;
    gchar *obj_path = NULL;
    GError *error = NULL;

    g_return_val_if_fail (tc->priv->state == TAPIOCA_CONNECTION_STATE_CONNECTED, NULL);

    if (!(org_tapioca_Connection_create_session (tc->priv->proxy,
                    (const guint) session_type,
                    &obj_path, 
                    &error))) {
        g_warning ("failed to create session");
        // TODO treat error as apropriate
        return NULL;
    }
    
    if (session_type == TAPIOCA_SESSION_TYPE_AUDIO)
        ts = TAPIOCA_SESSION (tapioca_audio_session_new (tc->priv->bus, tc->priv->service_name, obj_path));
    else if (session_type == TAPIOCA_SESSION_TYPE_TEXT)
        ts = TAPIOCA_SESSION (tapioca_text_session_new (tc->priv->bus, tc->priv->service_name, obj_path));

    if (obj_path) 
        g_free (obj_path);

    return ts;
}

/**
 * Returns a #TapiocaPresence instance for this connection.
 * Note: Do not delete this object.
 *
 * @param tc #TapiocaConnection instance.
 * @return #TapiocaPresence instance or NULL if fail.
 */
TapiocaPresence *
tapioca_connection_get_presence (TapiocaConnection *tc)
{
    g_return_val_if_fail (tc != NULL, NULL);

    return g_object_ref (tc->priv->presence);
}

/**
 * Returns a #TapiocaContactList instance for this connection.
 *
 * @param tc #TapiocaConnection instance.
 * @return #TapiocaContactList instance or NULL if fail.
 */
TapiocaContactList *
tapioca_connection_get_contact_list (TapiocaConnection *tc)
{
    g_return_val_if_fail (tc != NULL, NULL);
    
    return g_object_ref (tc->priv->contact_list);
}

G_CONST_RETURN gchar *
tapioca_connection_get_id (TapiocaConnection *tc)
{
    g_return_val_if_fail (tc != NULL, NULL);
    return tc->priv->id;
}

static void
destroy_connection (TapiocaConnection *tc)
{
    GError *error = NULL;

    dbus_g_proxy_disconnect_signal (tc->priv->proxy, "StateChanged",
            G_CALLBACK (proxy_state_changed_signal), tc);
    dbus_g_proxy_disconnect_signal (tc->priv->proxy, "IncomingSession",
            G_CALLBACK (proxy_incoming_session_signal), tc);
    dbus_g_proxy_disconnect_signal (tc->priv->proxy, "Error",
            G_CALLBACK (proxy_error_signal), tc);
   
    if (!(org_tapioca_Connection_destroy (tc->priv->proxy, &error))) {
        g_warning ("error destroying connection");
    }
}

static void
proxy_state_changed_signal (DBusGProxy *proxy,
                            guint state,
                            gpointer data)
{
    TapiocaConnection *tc = TAPIOCA_CONNECTION (data);
    
    tc->priv->state = state;

    g_signal_emit (tc, tapioca_connection_signals[STATE_CHANGED], 0, state);
}

static void
proxy_incoming_session_signal (DBusGProxy *proxy,
                               const gchar *from,
                               guint type,
                               const gchar *obj_path,
                               gpointer data)
{
    TapiocaConnection *tc = TAPIOCA_CONNECTION (data);
    TapiocaSession *session;

    if (type == TAPIOCA_SESSION_TYPE_AUDIO)
        session = TAPIOCA_SESSION (tapioca_audio_session_new (tc->priv->bus, tc->priv->service_name, obj_path));
    else if (type == TAPIOCA_SESSION_TYPE_TEXT)
        session = TAPIOCA_SESSION (tapioca_text_session_new (tc->priv->bus, tc->priv->service_name, obj_path));
    else {
        g_warning ("received incoming session signal with invalid session type"); 
        return;
    }

    g_signal_emit (tc, tapioca_connection_signals[INCOMING_SESSION], 0, from, type, session);
}

static void
proxy_error_signal (DBusGProxy *proxy,
                    guint code,
                    gpointer data)
{
    g_signal_emit (TAPIOCA_CONNECTION (data), tapioca_connection_signals[ERROR], 0, code);
}
