/*
 * Tapioca library
 * Copyright (C) 2005 INdT 
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapioca_connection_manager.h"

#include "tapiocad_connection_manager_bindings.h"
#include "tapioca_connection_private.h"
#include "tapioca_marshal.h"

#include <dbus/dbus-glib-bindings.h>

struct _TapiocaConnectionManagerPrivate {
    DBusGConnection *bus;
    DBusGProxy *proxy;
    gboolean disposed;
};

static void tapioca_connection_manager_dispose (GObject *object);

G_DEFINE_TYPE(TapiocaConnectionManager, tapioca_connection_manager, G_TYPE_OBJECT)

static gboolean tapioca_initialized = FALSE;

static void 
tapioca_connection_manager_class_init (TapiocaConnectionManagerClass *klass) 
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_connection_manager_parent_class = g_type_class_peek_parent (klass);

    g_type_class_add_private (klass, sizeof (TapiocaConnectionManagerPrivate));

    gobject_class->dispose = tapioca_connection_manager_dispose;
}

static void
tapioca_connection_manager_init (TapiocaConnectionManager *tcm)
{
    tcm->priv = TAPIOCA_CONNECTION_MANAGER_GET_PRIVATE (tcm);
    tcm->priv->bus = NULL;
    tcm->priv->proxy = NULL;
}

static void 
tapioca_connection_manager_dispose (GObject *object) 
{
    TapiocaConnectionManager *tp = TAPIOCA_CONNECTION_MANAGER (object);

    if (tp->priv->disposed) 
       /* If dispose did already run, return. */
       return;

    /* Make sure dispose does not run twice. */
    tp->priv->disposed = TRUE;
       
    g_object_unref (tp->priv->proxy);

    G_OBJECT_CLASS (tapioca_connection_manager_parent_class)->dispose (object);
}
        
/**
 * Returns a new #TapiocaConnectionManager instance.
 *
 * @return #TapiocaConnectionManager instance or NULL if fail.
 */
TapiocaConnectionManager *
tapioca_connection_manager_new (void) 
{
    TapiocaConnectionManager *tp;
    DBusGConnection *bus;
    DBusGProxy *proxy;
    GError *error = NULL;
 
    if (!tapioca_initialized) {
        tapioca_initialized = TRUE;

        /* register our own marshallers */
        dbus_g_object_register_marshaller (tapioca_marshal_VOID__UINT_STRING,
                G_TYPE_NONE, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_INVALID);
        dbus_g_object_register_marshaller (tapioca_marshal_VOID__STRING_UINT_STRING,
                G_TYPE_NONE, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_INVALID);
        dbus_g_object_register_marshaller (tapioca_marshal_VOID__STRING_UINT_STRING_BOOLEAN,
                G_TYPE_NONE, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_BOOLEAN, G_TYPE_INVALID);
        dbus_g_object_register_marshaller (tapioca_marshal_VOID__STRING_UINT_STRING_STRING_UINT,
                G_TYPE_NONE, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_INVALID);
        dbus_g_object_register_marshaller (tapioca_marshal_VOID__STRING_UINT_STRING_STRING_UINT_UINT,
                G_TYPE_NONE, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_UINT, G_TYPE_INVALID);
        dbus_g_object_register_marshaller (tapioca_marshal_VOID__STRING_STRING_STRING_UINT,
                G_TYPE_NONE, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_INVALID);
        dbus_g_object_register_marshaller (tapioca_marshal_VOID__STRING_POINTER,
                G_TYPE_NONE, G_TYPE_STRING, G_TYPE_POINTER, G_TYPE_INVALID);
        dbus_g_object_register_marshaller (tapioca_marshal_VOID__STRING_STRING_STRING_UINT_DOUBLE_STRING_STRING_STRING_UINT,
                G_TYPE_NONE, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_DOUBLE, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_INVALID);
    }
     
    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
    if (!bus) {
        g_warning ("failed to open connection to dbus");
        return NULL;
    }
        
    proxy = dbus_g_proxy_new_for_name (bus,
            "org.tapioca.Server",
            "/org/tapioca/ConnectionManager",
            "org.tapioca.ConnectionManager");
    if (!proxy) {
        g_warning ("failed to create proxy for service org.tapioca.Server with object path /org/tapioca/ConnectionManager");
        return NULL;
    }

    tp = TAPIOCA_CONNECTION_MANAGER (g_object_new (tapioca_connection_manager_get_type (), NULL));
    tp->priv->bus = bus;
    tp->priv->proxy = proxy;

    return tp;
}

// TODO implement get_supported_codecs

/**
 * Create a new #TapiocaConnection with state disconnected.
 *
 * @param tp #TapiocaConnectionManager instance.
 * @param protocol protocol that will be used..
 * @return #TapiocaConnection instance or NULL if fail.
 */
TapiocaConnection *
tapioca_connection_manager_create_connection (TapiocaConnectionManager *tp, 
                                              const gchar *protocol)
{
    TapiocaConnection *tc;
    gchar *service_name = NULL;
    gchar *obj_path = NULL;
    GError *error = NULL;

    /* Create a new connection with daemon */
    if (!(org_tapioca_ConnectionManager_create_connection (tp->priv->proxy, 
                    protocol, 
                    &service_name, 
                    &obj_path, 
                    &error))) {
        g_warning ("failed to create connection");
        return NULL;
    }

    tc = tapioca_connection_new (tp->priv->bus, service_name, obj_path);
    
    if (service_name)
        g_free (service_name);
    if (obj_path)
        g_free (obj_path);

    return tc;
}
