/*
 * Tapioca library
 * Copyright (C) 2005 INdT 
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapioca_contact_list.h"

#include "tapioca_contact_list_private.h"
#include "tapioca_contact_list_bindings.h"
#include "tapioca_marshal.h"

/* Signals */
enum {
    CONTACT_AUTHORIZED,
    CONTACT_DENIED,
    AUTHORIZATION_REQUESTED,
    CONTACTS_RETRIEVED,
    CONTACT_INFO_RETRIEVED,
    LAST_SIGNAL
};

/* Private members */
struct _TapiocaContactListPrivate {
    DBusGProxy *proxy;
    gboolean disposed;
};

static void tapioca_contact_list_dispose        (GObject *object);
static void tapioca_contact_list_finalize       (GObject *object);

static void proxy_contact_authorized_signal     (DBusGProxy *proxy,  
                                                 const gchar *addr,
                                                 gpointer data);
static void proxy_contact_denied_signal         (DBusGProxy *proxy, 
                                                 const gchar *addr,
                                                 gpointer data);
static void proxy_contact_auth_requested_signal (DBusGProxy *proxy,  
                                                 const gchar *addr, 
                                                 gpointer data);
static void proxy_contacts_retrieved_signal     (DBusGProxy *proxy,  
                                                 const gchar *addr,
                                                 const gchar *alias,
                                                 const gchar *group,
                                                 guint state,
                                                 gpointer data);
static void proxy_contact_info_retrieved_signal (DBusGProxy  *proxy,  
                                                 const gchar *addr, 
                                                 GHashTable *info,
                                                 gpointer data);

G_DEFINE_TYPE(TapiocaContactList, tapioca_contact_list, G_TYPE_OBJECT)
    
static guint tapioca_contact_list_signals[LAST_SIGNAL] = { 0 };

static void 
tapioca_contact_list_class_init (TapiocaContactListClass *klass) 
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_contact_list_parent_class = g_type_class_peek_parent (klass);
    
    g_type_class_add_private (klass, sizeof (TapiocaContactListPrivate));
    
    gobject_class->dispose = tapioca_contact_list_dispose;
    gobject_class->finalize = tapioca_contact_list_finalize;
    
    /* Let's create our signals */
    tapioca_contact_list_signals[CONTACT_AUTHORIZED] = g_signal_new ("contact-authorized",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__STRING,
        G_TYPE_NONE,
        1,
        G_TYPE_STRING);
    tapioca_contact_list_signals[CONTACT_DENIED] = g_signal_new ("contact-denied",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__STRING,
        G_TYPE_NONE,
        1,
        G_TYPE_STRING);
    tapioca_contact_list_signals[AUTHORIZATION_REQUESTED] = g_signal_new ("authorization-requested",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__STRING,
        G_TYPE_NONE,
        1,
        G_TYPE_STRING);
    tapioca_contact_list_signals[CONTACTS_RETRIEVED] = g_signal_new ("contacts-retrieved",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        tapioca_marshal_VOID__STRING_STRING_STRING_UINT,
        G_TYPE_NONE,
        4,
        G_TYPE_STRING,
        G_TYPE_STRING,
        G_TYPE_STRING,
        G_TYPE_UINT);
    tapioca_contact_list_signals[CONTACT_INFO_RETRIEVED] = g_signal_new ("contact-info-retrieved",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        tapioca_marshal_VOID__STRING_POINTER,
        G_TYPE_NONE,
        2,
        G_TYPE_STRING,
        G_TYPE_POINTER);
}

static void 
tapioca_contact_list_init (TapiocaContactList *tcl) 
{
    tcl->priv = TAPIOCA_CONTACT_LIST_GET_PRIVATE (tcl);
    tcl->priv->disposed = FALSE;
    tcl->priv->proxy = NULL;
}

static void 
tapioca_contact_list_dispose (GObject *object) 
{
    TapiocaContactList *tcl = TAPIOCA_CONTACT_LIST (object);

    if (tcl->priv->disposed) {
        /* If dispose did already run, return. */
        return;
    }
   
    /* Make sure dispose does not run twice. */
    tcl->priv->disposed = TRUE;

    /* Disconnect all signals */
    if (tcl->priv->proxy) {
        dbus_g_proxy_disconnect_signal (tcl->priv->proxy, "ContactAuthorized",
                G_CALLBACK (proxy_contact_authorized_signal), NULL);
        dbus_g_proxy_disconnect_signal (tcl->priv->proxy, "ContactDenied",
                G_CALLBACK (proxy_contact_denied_signal), NULL);
        dbus_g_proxy_disconnect_signal (tcl->priv->proxy, "AuthorizationRequested",
                G_CALLBACK (proxy_contact_auth_requested_signal), NULL);
        dbus_g_proxy_disconnect_signal (tcl->priv->proxy, "ContactsRetrieved",
                G_CALLBACK (proxy_contacts_retrieved_signal), NULL);
        dbus_g_proxy_disconnect_signal (tcl->priv->proxy, "ContactInfoRetrieved",
                G_CALLBACK (proxy_contact_info_retrieved_signal), NULL);
        
        g_object_unref (tcl->priv->proxy);
    }

    G_OBJECT_CLASS (tapioca_contact_list_parent_class)->dispose (object);
}

static void 
tapioca_contact_list_finalize (GObject *object) 
{
    g_signal_handlers_destroy (object);
    
    G_OBJECT_CLASS (tapioca_contact_list_parent_class)->finalize (object);
}

TapiocaContactList * 
tapioca_contact_list_new (DBusGConnection *bus,
                          const gchar *service_name,
                          const gchar *obj_path) 
{

    TapiocaContactList *tcl;
    DBusGProxy *proxy;

    proxy = dbus_g_proxy_new_for_name (bus,
            service_name,
            obj_path,
            "org.tapioca.ContactList");
    if (!proxy) {
        g_warning ("failed to create proxy for service %s with object path %s on interface org.tapioca.ContactList", 
                service_name, obj_path);
        return NULL;
    }

    tcl = TAPIOCA_CONTACT_LIST (g_object_new (tapioca_contact_list_get_type (), NULL));
    tcl->priv->proxy = proxy;

    /* Connect all contact list signals */
    dbus_g_proxy_add_signal (tcl->priv->proxy, "ContactAuthorized", G_TYPE_STRING, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (tcl->priv->proxy, "ContactDenied", G_TYPE_STRING, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (tcl->priv->proxy, "AuthorizationRequested", G_TYPE_STRING, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (tcl->priv->proxy, "ContactsRetrieved", G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (tcl->priv->proxy, "ContactInfoRetrieved", G_TYPE_STRING, G_TYPE_POINTER, G_TYPE_INVALID);
    
    dbus_g_proxy_connect_signal (tcl->priv->proxy, "ContactAuthorized",
            G_CALLBACK (proxy_contact_authorized_signal),
            tcl, NULL);
    dbus_g_proxy_connect_signal (tcl->priv->proxy, "ContactDenied",
            G_CALLBACK (proxy_contact_denied_signal),
            tcl, NULL);
    dbus_g_proxy_connect_signal (tcl->priv->proxy, "AuthorizationRequested",
            G_CALLBACK (proxy_contact_auth_requested_signal),
            tcl, NULL);
    dbus_g_proxy_connect_signal (tcl->priv->proxy, "ContactsRetrieved",
            G_CALLBACK (proxy_contacts_retrieved_signal),
            tcl, NULL);
    dbus_g_proxy_connect_signal (tcl->priv->proxy, "ContactInfoRetrieved",
            G_CALLBACK (proxy_contact_info_retrieved_signal),
            tcl, NULL);

    return tcl;
}

gboolean
tapioca_contact_list_request_contacts (TapiocaContactList *tcl)
{
    GError *error = NULL;

    g_return_val_if_fail (tcl != NULL, FALSE);

    if (!(org_tapioca_ContactList_request_contacts (tcl->priv->proxy, &error))) {
        g_warning ("failed to resquest contact list");
        return FALSE;
    }
    
    return TRUE;
}

gboolean
tapioca_contact_list_request_contact_info (TapiocaContactList *tcl,
                                           const gchar *addr)
{
    GError *error = NULL;

    g_return_val_if_fail (tcl != NULL, FALSE);
    g_return_val_if_fail (addr != NULL, FALSE);

    if (!(org_tapioca_ContactList_request_contact_info (tcl->priv->proxy, addr, &error))) {
        g_warning ("failed to resquest contact info");
        return FALSE;
    }
    
    return TRUE;
}

gboolean
tapioca_contact_list_add_contact (TapiocaContactList *tcl,
                                  const gchar *addr,
                                  const gchar *alias,
                                  const gchar *group)
{
    GError *error = NULL;

    g_return_val_if_fail (tcl != NULL, FALSE);
    g_return_val_if_fail (addr != NULL, FALSE);
    g_return_val_if_fail (alias != NULL, FALSE);
    g_return_val_if_fail (group != NULL, FALSE);

    if (!(org_tapioca_ContactList_add_contact (tcl->priv->proxy, addr, alias, group, &error))) {
        g_warning ("failed to add contact on contact list");
        return FALSE;
    }

    return TRUE;
}

gboolean
tapioca_contact_list_remove_contact (TapiocaContactList *tcl,
                                     const gchar *addr)
{
    GError *error = NULL;

    g_return_val_if_fail (tcl != NULL, FALSE);
    g_return_val_if_fail (addr != NULL, FALSE);

    if (!(org_tapioca_ContactList_remove_contact (tcl->priv->proxy, addr, &error))) {
        g_warning ("failed to remove contact");
        return FALSE;
    }

    return TRUE;
}

gboolean
tapioca_contact_list_authorize_contact (TapiocaContactList *tcl,
                                        const gchar *addr,
                                        const gchar *group)
{
    GError *error = NULL;

    g_return_val_if_fail (tcl != NULL, FALSE);
    g_return_val_if_fail (addr != NULL, FALSE);
    g_return_val_if_fail (group != NULL, FALSE);

    if (!(org_tapioca_ContactList_authorize_contact (tcl->priv->proxy, addr, group, &error))) {
        g_warning ("failed to authorize contact");
        return FALSE;
    }
    
    return TRUE;
}

gboolean
tapioca_contact_list_deny_contact (TapiocaContactList *tcl,
                                   const gchar *addr)
{
    GError *error = NULL;

    g_return_val_if_fail (tcl != NULL, FALSE);
    g_return_val_if_fail (addr != NULL, FALSE);

    if (!(org_tapioca_ContactList_deny_contact (tcl->priv->proxy, addr, &error))) {
        g_warning ("failed to deny contact.");
        return FALSE;
    }

    return TRUE;
}

static void 
proxy_contact_authorized_signal (DBusGProxy *proxy, 
                                 const gchar *addr,
                                 gpointer data)
{
    g_signal_emit (TAPIOCA_CONTACT_LIST (data), tapioca_contact_list_signals[CONTACT_AUTHORIZED], 0, addr);
}

static void 
proxy_contact_denied_signal (DBusGProxy *proxy, 
                             const gchar *addr,
                             gpointer data)
{
    g_signal_emit (TAPIOCA_CONTACT_LIST (data), tapioca_contact_list_signals[CONTACT_DENIED], 0, addr);
}

static void 
proxy_contact_auth_requested_signal (DBusGProxy *proxy, 
                                     const gchar *addr,
                                     gpointer data)
{
    g_signal_emit (TAPIOCA_CONTACT_LIST (data), tapioca_contact_list_signals[AUTHORIZATION_REQUESTED], 0, addr);
}

static void 
proxy_contacts_retrieved_signal (DBusGProxy *proxy,  
                                 const gchar *addr,
                                 const gchar *alias,
                                 const gchar *group,
                                 guint state,
                                 gpointer data)
{
    g_signal_emit (TAPIOCA_CONTACT_LIST (data), tapioca_contact_list_signals[CONTACTS_RETRIEVED], 0, addr, alias, group, state);
}

static void 
proxy_contact_info_retrieved_signal (DBusGProxy *proxy,  
                                     const gchar *addr, 
                                     GHashTable *info,
                                     gpointer data)
{
    g_signal_emit (TAPIOCA_CONTACT_LIST (data), tapioca_contact_list_signals[CONTACT_INFO_RETRIEVED], 0, addr, (gpointer) info);
}
