/*
 * Tapioca library
 * Copyright (C) 2005 INdT 
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>
#include "tapioca_presence.h"
#include "tapioca_marshal.h"
#include "tapioca_presence_bindings.h"

/* Signals */
enum {
    STATUS_UPDATED,
    REGISTERED,
    UNREGISTERED,
    LAST_SIGNAL
};

/* Private members */
struct _TapiocaPresencePrivate {
    gboolean disposed;
    DBusGProxy *proxy;
};

static void tapioca_presence_dispose        (GObject *object);
static void tapioca_presence_finalize       (GObject *object);

static void proxy_registered_signal         (DBusGProxy *proxy,  
                                             gpointer data);
static void proxy_unregistered_signal       (DBusGProxy *proxy, 
                                             gpointer data);
static void proxy_status_updated_signal     (DBusGProxy *proxy,  
                                             const gchar *contact, 
                                             guint status,
                                             const gchar *status_desc,
                                             gboolean voice_cap,
                                             gpointer data);

G_DEFINE_TYPE(TapiocaPresence, tapioca_presence, G_TYPE_OBJECT)
   
static guint tapioca_presence_signals[LAST_SIGNAL] = { 0 };

static void 
tapioca_presence_class_init (TapiocaPresenceClass *klass) 
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_presence_parent_class = g_type_class_peek_parent (klass);
    
    g_type_class_add_private (klass, sizeof (TapiocaPresencePrivate));
    
    gobject_class->dispose = tapioca_presence_dispose;
    gobject_class->finalize = tapioca_presence_finalize;
    
    /* Let's create our signals */
    tapioca_presence_signals[STATUS_UPDATED] = g_signal_new ("status-updated",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        tapioca_marshal_VOID__STRING_UINT_STRING_BOOLEAN,
        G_TYPE_NONE,
        4,
        G_TYPE_STRING,
        G_TYPE_UINT,
        G_TYPE_STRING,
        G_TYPE_BOOLEAN);
    tapioca_presence_signals[REGISTERED] = g_signal_new ("registered",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__VOID,
        G_TYPE_NONE,
        0);
    tapioca_presence_signals[UNREGISTERED] = g_signal_new ("unregistered",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__VOID,
        G_TYPE_NONE,
        0);
}

static void 
tapioca_presence_init (TapiocaPresence *tp) 
{
    tp->priv = TAPIOCA_PRESENCE_GET_PRIVATE (tp);
    tp->priv->proxy = NULL;
    tp->priv->disposed = FALSE;
}

static void 
tapioca_presence_dispose (GObject *object) 
{
    TapiocaPresence *tp = TAPIOCA_PRESENCE (object);

    if (tp->priv->disposed) {
        /* If dispose did already run, return. */
        return;
    }
   
    /* Make sure dispose does not run twice. */
    tp->priv->disposed = TRUE;

    /* Disconnect all signals */
    if (tp->priv->proxy) {
        dbus_g_proxy_disconnect_signal (tp->priv->proxy, "StatusUpdated",
                G_CALLBACK (proxy_status_updated_signal), NULL);
        dbus_g_proxy_disconnect_signal (tp->priv->proxy, "Registered",
                G_CALLBACK (proxy_registered_signal), NULL);
        dbus_g_proxy_disconnect_signal (tp->priv->proxy, "Unregistered",
                G_CALLBACK (proxy_unregistered_signal), NULL);

        g_object_unref (tp->priv->proxy);
    }

    G_OBJECT_CLASS (tapioca_presence_parent_class)->dispose (object);
}

static void 
tapioca_presence_finalize (GObject *object) 
{
    g_signal_handlers_destroy (object);
    
    G_OBJECT_CLASS (tapioca_presence_parent_class)->finalize (object);
}

TapiocaPresence * 
tapioca_presence_new (DBusGConnection *bus,
                      const gchar *service_name, 
                      const gchar *obj_path)
{
    TapiocaPresence *tp;
    DBusGProxy *proxy;

    /* Request a new proxy for tp service */
    proxy = dbus_g_proxy_new_for_name (bus,
            service_name,
            obj_path,
            "org.tapioca.Presence");
    if (!proxy) {
        g_warning ("failed to create proxy for service %s with object path %s on interface org.tapioca.Presence", 
                service_name, obj_path);
        return NULL;
    }

    tp = TAPIOCA_PRESENCE (g_object_new (tapioca_presence_get_type (), NULL));
    tp->priv->proxy = proxy;

    /* Connect all tp signals */
    dbus_g_proxy_add_signal (tp->priv->proxy, "StatusUpdated", G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_BOOLEAN, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (tp->priv->proxy, "Unregistered", G_TYPE_INVALID);
    dbus_g_proxy_add_signal (tp->priv->proxy, "Registered", G_TYPE_INVALID);

    dbus_g_proxy_connect_signal (tp->priv->proxy, "StatusUpdated",
            G_CALLBACK (proxy_status_updated_signal),
            tp, NULL);
    dbus_g_proxy_connect_signal (tp->priv->proxy, "Registered",
            G_CALLBACK (proxy_registered_signal),
            tp, NULL);
    dbus_g_proxy_connect_signal (tp->priv->proxy, "Unregistered",
            G_CALLBACK (proxy_unregistered_signal),
            tp, NULL);

    return tp;
}

gboolean
tapioca_presence_subscribe (TapiocaPresence *tp,
                            TapiocaPresenceStatus status,
                            const char *status_desc)
{
    GError *error = NULL;

    g_return_val_if_fail (tp != NULL, FALSE);

    if (!(org_tapioca_Presence_subscribe (tp->priv->proxy, (const guint) status, status_desc, &error))) {
        g_warning ("failed to subscribe presence");
        return FALSE;
    }
    
    return TRUE;
}

gboolean 
tapioca_presence_unsubscribe (TapiocaPresence *tp)
{
    GError *error = NULL;

    g_return_val_if_fail (tp != NULL, FALSE);

    if (!(org_tapioca_Presence_unsubscribe (tp->priv->proxy, &error))) {
        g_warning ("failed to unsubscribe presence");
        return FALSE;
    }

    return TRUE;
}

GArray *
tapioca_presence_get_supported_statuses (TapiocaPresence *tp)
{
    GError *error = NULL;
    GArray *statuses = NULL;

    g_return_val_if_fail (tp != NULL, NULL);

    if (!(org_tapioca_Presence_get_supported_statuses (tp->priv->proxy, &statuses, &error))) {
        g_warning ("failed to get supported presence statuses");
        return NULL;
    }
    
    return statuses;
}

gboolean
tapioca_presence_set_status (TapiocaPresence *tp,
                             TapiocaPresenceStatus status,
                             const char *status_desc)
{
    GError *error = NULL;

    g_return_val_if_fail (tp != NULL, FALSE);

    if (!(org_tapioca_Presence_set_status (tp->priv->proxy, (const guint) status, status_desc, &error))) {
        g_warning ("failed to set presence status");
        return FALSE;
    }
    
    return TRUE;
}

gboolean 
tapioca_presence_request_status (TapiocaPresence *tp,
                                 const gchar **list)
{
    GError *error = NULL;

    g_return_val_if_fail (tp != NULL, FALSE);

    if (!(org_tapioca_Presence_request_status (tp->priv->proxy, list, &error))) {
        g_warning ("failed to resquest presence status");
        return FALSE;
    }

    return TRUE;
}

gboolean
tapioca_presence_release_status (TapiocaPresence *tp,
                                 const gchar **list)
{
    GError *error = NULL;

    g_return_val_if_fail (tp != NULL, FALSE);

    if (!(org_tapioca_Presence_release_status (tp->priv->proxy, list, &error))) {
        g_warning ("failed to release presence status");
        return FALSE;
    }

    return TRUE;
}

static void 
proxy_status_updated_signal (DBusGProxy *proxy, 
                             const gchar *contact,
                             guint status, 
                             const gchar *status_desc,
                             gboolean voice_cap,
                             gpointer data)
{
    g_signal_emit (TAPIOCA_PRESENCE (data), tapioca_presence_signals[STATUS_UPDATED], 0, contact, status, status_desc, voice_cap);
}

static void 
proxy_registered_signal (DBusGProxy *proxy, 
                         gpointer data)
{
    g_signal_emit (TAPIOCA_PRESENCE (data), tapioca_presence_signals[REGISTERED], 0);
}

static void 
proxy_unregistered_signal (DBusGProxy *proxy, 
                           gpointer data)
{
    g_signal_emit (TAPIOCA_PRESENCE (data), tapioca_presence_signals[UNREGISTERED], 0);
}
