/*
 * Tapioca library
 * Copyright (C) 2005 INdT 
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapioca_session.h"

#include "tapioca_session_private.h"
#include "tapioca_session_bindings.h"
#include "tapioca_marshal.h"

#include <dbus/dbus-glib-bindings.h>

/* Signals */
enum {
    STATE_CHANGED,
    ERROR,
    MEMBER_ADDED,
    MEMBER_REMOVED,
    ADD_MEMBER_ERROR,
    LAST_SIGNAL
};

/* Private members */
struct _TapiocaSessionPrivate {
    DBusGProxy *proxy;
    TapiocaSessionState state;
    gboolean disposed;
};

static void tapioca_session_dispose       (GObject *object);
static void tapioca_session_finalize      (GObject *object);

static void destroy_session               (TapiocaSession *ts);
static void proxy_state_changed_signal    (DBusGProxy *proxy, 
                                           guint state,
                                           gpointer data);
static void proxy_member_added_signal     (DBusGProxy *proxy, 
                                           const gchar *member,
                                           gpointer data);
static void proxy_member_removed_signal   (DBusGProxy *proxy, 
                                           const gchar *member,  
                                           gpointer data);
static void proxy_add_member_error_signal (DBusGProxy *proxy, 
                                           const gchar *member, 
                                           gpointer data);
static void proxy_error_signal            (DBusGProxy *proxy, 
                                           guint code,
                                           gpointer data);

G_DEFINE_TYPE(TapiocaSession, tapioca_session, G_TYPE_OBJECT)
    
static guint tapioca_session_signals[LAST_SIGNAL] = { 0 };

static void 
tapioca_session_class_init (TapiocaSessionClass *klass) 
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_session_parent_class = g_type_class_peek_parent (klass);
    
    g_type_class_add_private (klass, sizeof (TapiocaSessionPrivate));
    
    gobject_class->dispose = tapioca_session_dispose;
    gobject_class->finalize = tapioca_session_finalize;

    /* Let's create our signals */
    tapioca_session_signals[STATE_CHANGED] = g_signal_new ("state-changed",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__UINT,
        G_TYPE_NONE,
        1,
        G_TYPE_UINT);
    tapioca_session_signals[MEMBER_ADDED] = g_signal_new ("member-added",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__STRING,
        G_TYPE_NONE,
        1,
        G_TYPE_STRING);
    tapioca_session_signals[MEMBER_REMOVED] = g_signal_new ("member-removed",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__STRING,
        G_TYPE_NONE,
        1,
        G_TYPE_STRING);
    tapioca_session_signals[ADD_MEMBER_ERROR] = g_signal_new ("add-member-error",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__UINT,
        G_TYPE_NONE,
        1,
        G_TYPE_UINT);
    tapioca_session_signals[ERROR] = g_signal_new ("error",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        g_cclosure_marshal_VOID__UINT,
        G_TYPE_NONE,
        1,
        G_TYPE_UINT);
}

static void 
tapioca_session_init (TapiocaSession *ts) 
{
    ts->priv = TAPIOCA_SESSION_GET_PRIVATE (ts);
    ts->priv->proxy = NULL;
    ts->priv->state = TAPIOCA_SESSION_STATE_CLOSED;
    ts->priv->disposed = FALSE;
}

static void 
tapioca_session_dispose (GObject *object) 
{
    TapiocaSession *ts = TAPIOCA_SESSION (object);

    if (ts->priv->disposed) {
        /* If dispose did already run, return. */
        return;
    }
   
    /* Make sure dispose does not run twice. */
    ts->priv->disposed = TRUE;

    if (ts->priv->proxy) {
        dbus_g_proxy_disconnect_signal (ts->priv->proxy, "StateChanged", 
                G_CALLBACK (proxy_state_changed_signal), NULL);
        dbus_g_proxy_disconnect_signal (ts->priv->proxy, "Error",
                G_CALLBACK (proxy_error_signal), NULL);
        dbus_g_proxy_disconnect_signal (ts->priv->proxy, "MemberAdded",
                G_CALLBACK (proxy_member_added_signal), NULL);
        dbus_g_proxy_disconnect_signal (ts->priv->proxy, "MemberRemoved",
                G_CALLBACK (proxy_member_removed_signal), NULL);
        dbus_g_proxy_disconnect_signal (ts->priv->proxy, "AddMemberError",
                G_CALLBACK (proxy_add_member_error_signal), NULL);

        if (ts->priv->state != TAPIOCA_SESSION_STATE_CLOSED) 
            destroy_session (ts);

        g_object_unref (ts->priv->proxy);
    }

    G_OBJECT_CLASS (tapioca_session_parent_class)->dispose (object);
}

static void 
tapioca_session_finalize (GObject *object) 
{
    g_signal_handlers_destroy (object);
    
    G_OBJECT_CLASS (tapioca_session_parent_class)->finalize (object);
}

gboolean
tapioca_session_setup (TapiocaSession *ts, 
                       DBusGConnection *bus,
                       const gchar *service_name,
                       const gchar *obj_path)
{
    DBusGProxy *proxy;

    proxy = dbus_g_proxy_new_for_name (bus,
            service_name,
            obj_path,
            "org.tapioca.Session");
    if (!proxy) {
        g_warning ("failed to create proxy for service %s with object path %s on interface org.tapioca.Session", 
                service_name, obj_path);
        return FALSE;
    }
    
    ts->priv->proxy = proxy;

    dbus_g_proxy_add_signal (ts->priv->proxy, "StateChanged", G_TYPE_UINT, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (ts->priv->proxy, "Error", G_TYPE_UINT, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (ts->priv->proxy, "MemberAdded", G_TYPE_STRING, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (ts->priv->proxy, "MemberRemoved", G_TYPE_STRING, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (ts->priv->proxy, "AddMemberError", G_TYPE_UINT, G_TYPE_STRING, G_TYPE_INVALID);
    
    dbus_g_proxy_connect_signal (ts->priv->proxy, "StateChanged",
            G_CALLBACK (proxy_state_changed_signal),
            ts, NULL);
    dbus_g_proxy_connect_signal (ts->priv->proxy, "MemberAdded",
            G_CALLBACK (proxy_member_added_signal),
            ts, NULL);
    dbus_g_proxy_connect_signal (ts->priv->proxy, "MemberRemoved",
            G_CALLBACK (proxy_member_removed_signal),
            ts, NULL);
    dbus_g_proxy_connect_signal (ts->priv->proxy, "AddMemberError",
            G_CALLBACK (proxy_add_member_error_signal),
            ts, NULL);
    dbus_g_proxy_connect_signal (ts->priv->proxy, "Error",
            G_CALLBACK (proxy_error_signal),
            ts, NULL);
    
    return TRUE;
}

/**
 * Initiate session.
 *
 * @param ts #TapiocaSession instance.
 * @return FALSE if an error occurs otherwise TRUE.
 */
gboolean  
tapioca_session_initiate (TapiocaSession *ts)
{
    GError *error = NULL;

    g_return_val_if_fail (ts != NULL, FALSE);
    g_return_val_if_fail (ts->priv->state == TAPIOCA_SESSION_STATE_CLOSED, FALSE);

    if (!(org_tapioca_Session_initiate (ts->priv->proxy, &error))) {
        g_warning ("failed to initiate session");
        return FALSE;
    }

    return TRUE;
}

/**
 * Close session.
 *
 * @param ts #TapiocaSession instance.
 */
void
tapioca_session_close (TapiocaSession *ts)
{
    GError *error = NULL;

    g_return_if_fail (ts != NULL);
    g_return_if_fail (ts->priv->state != TAPIOCA_SESSION_STATE_CLOSED);

    if (!(org_tapioca_Session_close (ts->priv->proxy, &error))) {
        g_warning ("failed to close session");
    }
}

/**
 * Return an array of members.
 *
 * @param ts #TapiocaSession instance.
 * @return Return an array with members or NULL if it fail.
 */
gchar ** 
tapioca_session_get_members  (TapiocaSession *ts)
{
    gchar **members = NULL;
    GError *error = NULL;
    
    g_return_val_if_fail (ts != NULL, NULL);

    if (!(org_tapioca_Session_get_members (ts->priv->proxy, &members, &error))) {
        g_warning ("failed to get members of session");
        return NULL;
    }

    return members;
}

/**
 * Add members to session.
 *
 * @param ts #TapiocaSession instance.
 * @param members array of members that will be add.
 * @return FALSE if an error occurs otherwise TRUE.
 */
gboolean
tapioca_session_add_members (TapiocaSession *ts, 
                             const gchar **members)
{
    GError *error = NULL;

    g_return_val_if_fail (ts != NULL, FALSE);
    g_return_val_if_fail (members != NULL, FALSE);

    if (!(org_tapioca_Session_add_members (ts->priv->proxy, members, &error))) {
        g_warning ("failed to add members of session");
        return FALSE;
    }

    return TRUE;
}

/** 
 * Remove members from session.
 *
 * @param ts #TapiocaSession instance.
 * @param members array of members that will be removed.
 * @return FALSE if an error occurs otherwise TRUE
 */
gboolean
tapioca_session_remove_members (TapiocaSession *ts, 
                                const gchar **members)
{
    GError *error = NULL;

    g_return_val_if_fail (ts != NULL, FALSE);
    g_return_val_if_fail (members != NULL, FALSE);

    if (!(org_tapioca_Session_remove_members (ts->priv->proxy, members, &error))) {
        g_warning ("failed to remove members of session");
        return FALSE;
    }

    return TRUE;
}

/**
 * Return the type of the session.
 *
 * @param ts #TapiocaSession instance
 * @return #TapiocaSessionType
 */
TapiocaSessionType
tapioca_session_get_session_type (TapiocaSession *ts)
{
    guint type;
    GError *error = NULL;

    g_return_val_if_fail (ts != NULL, TAPIOCA_SESSION_TYPE_UNKNOWN);

    if (!(org_tapioca_Session_get_type (ts->priv->proxy, &type, &error))) {
        g_warning ("failed to get the session type");
        return TAPIOCA_SESSION_TYPE_UNKNOWN;
    }

    return type;
}

/**
 * Return the current session state.
 *
 * @param ts #TapiocaSession instance
 * @return #TapiocaSessionState
 */
TapiocaSessionState 
tapioca_session_get_state (TapiocaSession *ts)
{
    return ts->priv->state;
}

static void
destroy_session (TapiocaSession *ts)
{
    GError *error = NULL;

    if (!org_tapioca_Session_destroy (ts->priv->proxy, &error)) {
        g_warning ("error destroying session");
    }
}

static void 
proxy_state_changed_signal (DBusGProxy *proxy, 
                            guint state, 
                            gpointer data)
{
    TapiocaSession *ts = TAPIOCA_SESSION (data);
    
    ts->priv->state = state;
    
    g_signal_emit (ts, tapioca_session_signals[STATE_CHANGED], 0, state);
}

static void 
proxy_member_added_signal (DBusGProxy *proxy, 
                           const gchar *member, 
                           gpointer data)
{
    g_signal_emit (TAPIOCA_SESSION (data), tapioca_session_signals[MEMBER_ADDED], 0, member);
}

static void 
proxy_member_removed_signal (DBusGProxy *proxy, 
                             const gchar *member, 
                             gpointer data)
{
    g_signal_emit (TAPIOCA_SESSION (data), tapioca_session_signals[MEMBER_REMOVED], 0, member);
}

static void 
proxy_add_member_error_signal (DBusGProxy *proxy,
                               const gchar *member, 
                               gpointer data)
{
    g_signal_emit (TAPIOCA_SESSION (data), tapioca_session_signals[ADD_MEMBER_ERROR], 0, member);
}

static void 
proxy_error_signal (DBusGProxy *proxy, 
                    guint code, 
                    gpointer data)
{
    g_signal_emit (TAPIOCA_SESSION (data), tapioca_session_signals[ERROR], 0, code);
}
