/*
 * Tapioca library
 * Copyright (C) 2005 INdT 
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapioca_text_session.h"

#include "tapioca_session_private.h"
#include "tapioca_text_session_private.h"
#include "tapioca_text_session_bindings.h"
#include "tapioca_marshal.h"

#include <dbus/dbus-glib-bindings.h>

/* Signals */
enum {
    INCOMING,
    DELIVERY_ERROR,
    LAST_SIGNAL
};

/* Private members */
struct _TapiocaTextSessionPrivate {
    DBusGProxy *proxy;
    gboolean disposed;
};

static void tapioca_text_session_dispose        (GObject *object);
static void tapioca_text_session_finalize       (GObject *object);

static void proxy_incoming                      (DBusGProxy *proxy, 
                                                 const gchar *from, 
                                                 guint type, 
                                                 const gchar *subject,  
                                                 const gchar *body,  
                                                 guint timestamp, 
                                                 gpointer data);
static void proxy_delivery_error                (DBusGProxy *proxy, 
                                                 const gchar *member,
                                                 guint type, 
                                                 const gchar *subject,
                                                 const gchar *body,
                                                 guint timestamp, 
                                                 guint error_code, 
                                                 gpointer data);

G_DEFINE_TYPE(TapiocaTextSession, tapioca_text_session, TAPIOCA_TYPE_SESSION)

static guint tapioca_text_session_signals[LAST_SIGNAL] = { 0 };

static void 
tapioca_text_session_class_init (TapiocaTextSessionClass *klass) 
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_text_session_parent_class = g_type_class_peek_parent (klass);

    g_type_class_add_private (klass, sizeof (TapiocaTextSessionPrivate));

    gobject_class->dispose = tapioca_text_session_dispose;
    gobject_class->finalize = tapioca_text_session_finalize;

    /* Create signals */
    tapioca_text_session_signals[INCOMING] = g_signal_new ("incoming",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        tapioca_marshal_VOID__STRING_UINT_STRING_STRING_UINT,
        G_TYPE_NONE,
        5,
        G_TYPE_STRING,
        G_TYPE_UINT,
        G_TYPE_STRING,
        G_TYPE_STRING,
        G_TYPE_UINT);
    tapioca_text_session_signals[DELIVERY_ERROR] = g_signal_new ("delivery-error",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
        0,
        NULL,
        NULL,
        tapioca_marshal_VOID__STRING_UINT_STRING_STRING_UINT_UINT,
        G_TYPE_NONE,
        6,
        G_TYPE_STRING,
        G_TYPE_UINT,
        G_TYPE_STRING,
        G_TYPE_STRING,
        G_TYPE_UINT,
        G_TYPE_UINT);
}

static void 
tapioca_text_session_init (TapiocaTextSession *tts) 
{
    tts->priv = TAPIOCA_TEXT_SESSION_GET_PRIVATE (tts);
    tts->priv->proxy = NULL;
    tts->priv->disposed = FALSE;
}

static void 
tapioca_text_session_dispose (GObject *object) 
{
    TapiocaTextSession *tts = TAPIOCA_TEXT_SESSION (object);

    if (tts->priv->disposed) {
        /* If dispose did already run, return. */
        return;
    }
   
    /* Make sure dispose does not run twice. */
    tts->priv->disposed = TRUE;

    if (tts->priv->proxy) {
        dbus_g_proxy_disconnect_signal (tts->priv->proxy, "Incoming",
                G_CALLBACK (proxy_incoming), NULL);
        dbus_g_proxy_disconnect_signal (tts->priv->proxy, "DeliveryError",
                G_CALLBACK (proxy_delivery_error), NULL);
    
        g_object_unref (tts->priv->proxy);
    }
 
    G_OBJECT_CLASS (tapioca_text_session_parent_class)->dispose (object);
}

static void 
tapioca_text_session_finalize (GObject *object) 
{
    g_signal_handlers_destroy (object);
    
    G_OBJECT_CLASS (tapioca_text_session_parent_class)->finalize (object);
}

TapiocaTextSession *
tapioca_text_session_new (DBusGConnection *bus,
                          const gchar *service_name, 
                          const gchar *obj_path) 
{
    TapiocaTextSession *tts;
    DBusGProxy *proxy;

    proxy = dbus_g_proxy_new_for_name (bus,
            service_name,
            obj_path,
            "org.tapioca.Session.Text");
    if (!proxy) {
        g_warning ("failed to create proxy for service %s with object path %s on interface org.tapioca.Session.Text", 
                service_name, obj_path);
        return NULL;
    }

    /* Create and set a new tapioca text instance */
    tts = TAPIOCA_TEXT_SESSION (g_object_new (tapioca_text_session_get_type (), NULL));
   
    if (!tapioca_session_setup (TAPIOCA_SESSION (tts), bus, service_name, obj_path)) {
        g_object_unref (tts);
        return NULL;
    }
    
    tts->priv->proxy = proxy;

    dbus_g_proxy_add_signal (tts->priv->proxy, "Incoming", 
            G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_INVALID);
    dbus_g_proxy_add_signal (tts->priv->proxy, "DeliveryError", 
            G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, G_TYPE_UINT, G_TYPE_INVALID);
    dbus_g_proxy_connect_signal (tts->priv->proxy, "Incoming",
            G_CALLBACK (proxy_incoming),
            tts, NULL);
    dbus_g_proxy_connect_signal (tts->priv->proxy, "DeliveryError",
            G_CALLBACK (proxy_delivery_error),
            tts, NULL);

    return tts;
}

/**
 * Send a text message.
 * This function can send a text plain message
 * or a control message (e.g. IS_TYPING_MESSAGE).
 *
 * @param tts #TapiocaTextSession instance.
 * @param type message type.
 * @param subject message subject.
 * @param body message body.
 * @return FALSE if an error occurs otherwise TRUE
 */
gboolean
tapioca_text_session_send (TapiocaTextSession *tts,
                           TapiocaTextSessionMessageType type,
                           const gchar *subject,
                           const gchar *body)
{
    GError *error = NULL;

    g_return_val_if_fail (tts != NULL, FALSE);

    if (!org_tapioca_Session_Text_send (
                tts->priv->proxy, (const guint) type, subject, body, &error)) {
        g_warning ("failed to send message");
        return FALSE;
    }

    return TRUE;
}

void
tapioca_text_session_clear_message_queue (TapiocaTextSession *tts)
{
    GError *error = NULL;
    
    g_return_if_fail (tts != NULL);

    if (!org_tapioca_Session_Text_clear_message_queue (
                tts->priv->proxy, &error)) {
        g_warning ("failed to clear message queue");
        // TODO treat error as apropriate
    }
}

void
tapioca_text_session_enable_message_queue (TapiocaTextSession *tts)
{
    GError *error = NULL;
    
    g_return_if_fail (tts != NULL);

    if (!org_tapioca_Session_Text_set_enable_message_queue (
                tts->priv->proxy, TRUE, &error)) {
        g_warning ("failed to enable message queue");
        // TODO treat error as apropriate
    }
}

void
tapioca_text_session_disable_message_queue (TapiocaTextSession *tts)
{
    GError *error = NULL;
    
    g_return_if_fail (tts != NULL);
    
    if (!org_tapioca_Session_Text_set_enable_message_queue (
                tts->priv->proxy, FALSE, &error)) {
        g_warning ("failed to disable message queue");
        // TODO treat error as apropriate
    }
}
                 
static void
proxy_incoming (DBusGProxy *proxy,
                const gchar *from,
                guint type,
                const gchar *subject,
                const gchar *body,
                guint timestamp,
                gpointer data)
{
    g_signal_emit (TAPIOCA_TEXT_SESSION (data), 
            tapioca_text_session_signals[INCOMING], 0, 
            from, type, subject, body, timestamp);
}

static void 
proxy_delivery_error (DBusGProxy *proxy, 
                      const gchar *member,
                      guint type, 
                      const gchar *subject,
                      const gchar *body,
                      guint timestamp,
                      guint error_code, 
                      gpointer data)
{
    g_signal_emit (TAPIOCA_TEXT_SESSION (data), 
            tapioca_text_session_signals[DELIVERY_ERROR], 0, 
            member, type, subject, body, timestamp, error_code);
}
