/*
 * Tapioca library
 * Copyright (C) 2005 INdT 
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapioca_audio_session.h"
#include "tapioca_audio_session_private.h"
#include "tapioca_audio_session_glue.h"

/* workaround: remove this when dbus works with child objects
   calling parent objects methods/signals */
#include "tapioca_session_private.h"

#include "tapioca_enums.h"
#include "tapioca_errors.h"
#include "tapioca_marshal.h"
#include "tapioca_utils.h"

#include <glib-object.h>
#include <dbus/dbus-glib-bindings.h>

enum {
    REMOTE_CAPABILITIES_UPDATED,
    REMOTE_CANDIDATE_ADDED,
    LAST_SIGNAL
};

struct _TapiocaAudioSessionPrivate 
{
    gboolean disposed;
};

G_DEFINE_TYPE(TapiocaAudioSession, tapioca_audio_session, TAPIOCA_TYPE_SESSION)
    
static void tapioca_audio_session_dispose  (GObject *object);
static void tapioca_audio_session_finalize (GObject *object);

static guint tapioca_audio_session_signals[LAST_SIGNAL] = { 0 };

static void
tapioca_audio_session_class_init (TapiocaAudioSessionClass *klass)
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_audio_session_parent_class = g_type_class_peek_parent (klass);

    g_type_class_add_private (klass, sizeof (TapiocaAudioSessionPrivate));
   
    gobject_class->dispose = tapioca_audio_session_dispose;
    gobject_class->finalize = tapioca_audio_session_finalize;
    
    klass->accept = NULL;
    klass->reject = NULL;
    klass->set_local_capabilities = NULL;
    klass->get_capabilities = NULL;
    klass->add_local_candidate = NULL;
    
    /* Let's create our signals */
    tapioca_audio_session_signals[REMOTE_CAPABILITIES_UPDATED] = g_signal_new ("remote-capabilities-updated",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            g_cclosure_marshal_VOID__STRING,
            G_TYPE_NONE,
            1,
            G_TYPE_STRING);
    tapioca_audio_session_signals[REMOTE_CANDIDATE_ADDED] = g_signal_new ("remote-candidate-added",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            tapioca_marshal_VOID__STRING_STRING_STRING_UINT_DOUBLE_STRING_STRING_STRING_UINT,
            G_TYPE_NONE,
            9,
            G_TYPE_STRING,
            G_TYPE_STRING,
            G_TYPE_STRING,
            G_TYPE_UINT,
            G_TYPE_DOUBLE,
            G_TYPE_STRING,
            G_TYPE_STRING,
            G_TYPE_STRING,
            G_TYPE_UINT);

    dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (klass), 
            &dbus_glib_tapioca_audio_session_object_info);
}

static void
tapioca_audio_session_init (TapiocaAudioSession *tas)
{
    TapiocaSession *ts = TAPIOCA_SESSION (tas);
    ts->session_type = TAPIOCA_SESSION_TYPE_AUDIO;
    
    tas->priv = TAPIOCA_AUDIO_SESSION_GET_PRIVATE (tas);
    tas->priv->disposed = FALSE;
}

static void
tapioca_audio_session_dispose (GObject *object)
{    
    TapiocaAudioSession *tas = TAPIOCA_AUDIO_SESSION (object);

    if (tas->priv->disposed)
        return;

    tas->priv->disposed = TRUE;
    
    G_OBJECT_CLASS (tapioca_audio_session_parent_class)->dispose (object);
}

static void
tapioca_audio_session_finalize (GObject *object)
{
    g_signal_handlers_destroy (object);
    
    G_OBJECT_CLASS (tapioca_audio_session_parent_class)->finalize (object);
}

void 
tapioca_audio_session_signal_remote_capabilities_updated (TapiocaAudioSession *tas,
                                                          const gchar *remote_capabilities)
{
    g_signal_emit (tas, tapioca_audio_session_signals[REMOTE_CAPABILITIES_UPDATED],
            0, remote_capabilities);
}

void
tapioca_audio_session_signal_remote_candidate_added (TapiocaAudioSession *tas, 
                                                     const gchar *name,
                                                     const gchar *protocol,
                                                     const gchar *ip,
                                                     guint port,
                                                     gdouble preference,
                                                     const gchar *username,
                                                     const gchar *password,
                                                     const gchar *type,
                                                     guint generation)
{
    g_signal_emit (tas, 
            tapioca_audio_session_signals[REMOTE_CANDIDATE_ADDED],
            0,
            name,
            protocol,
            ip,
            port,
            preference,
            username,
            password,
            type,
            generation);
}

gboolean
tapioca_audio_session_accept (TapiocaAudioSession *tas,
                              GError **error)
{
    TapiocaAudioSessionClass *klass = TAPIOCA_AUDIO_SESSION_GET_CLASS (tas);
    
    g_set_error_val_if_fail (klass->accept != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Not implemented");

    if (!klass->accept (tas, error)) {
        if (*error == NULL)
            g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, NULL);
        return FALSE;
    }

    return TRUE;
}

gboolean
tapioca_audio_session_reject (TapiocaAudioSession *tas,
                              GError **error)
{
    TapiocaAudioSessionClass *klass = TAPIOCA_AUDIO_SESSION_GET_CLASS (tas);
    
    g_set_error_val_if_fail (klass->reject != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Not implemented");

    if (!klass->reject (tas, error)) {
        if (*error == NULL)
            g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, NULL);
        return FALSE;
    }

    return TRUE;
}

gboolean
tapioca_audio_session_set_local_capabilities (TapiocaAudioSession *tas, 
                                              const gchar *local_capabilities,
                                              GError **error)
{
    TapiocaAudioSessionClass *klass = TAPIOCA_AUDIO_SESSION_GET_CLASS (tas);
    
    g_set_error_val_if_fail (klass->set_local_capabilities != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Not implemented");

    if (!klass->set_local_capabilities (tas, local_capabilities, error)) {
        if (*error == NULL)
            g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, NULL);
        return FALSE;
    }

    return TRUE;
}

gboolean
tapioca_audio_session_get_capabilities (TapiocaAudioSession *tas, 
                                        gchar **local_capabilities,
                                        gchar **remote_capabilities,
                                        GError **error)
{
    TapiocaAudioSessionClass *klass = TAPIOCA_AUDIO_SESSION_GET_CLASS (tas);
    
    g_set_error_val_if_fail (klass->get_capabilities != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Not implemented");

    if (!klass->get_capabilities (tas, local_capabilities, remote_capabilities, error)) {
        if (*error == NULL)
            g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, NULL);
        return FALSE;
    }
    
    return TRUE;
}

gboolean
tapioca_audio_session_add_local_candidate (TapiocaAudioSession *tas, 
                                           const gchar *name,
                                           const gchar *protocol,
                                           const gchar *ip,
                                           guint port,
                                           gdouble preference,
                                           const gchar *username,
                                           const gchar *password,
                                           const gchar *type,
                                           guint generation,
                                           GError **error)
{
    TapiocaAudioSessionClass *klass = TAPIOCA_AUDIO_SESSION_GET_CLASS (tas);
    
    g_set_error_val_if_fail (TAPIOCA_SESSION (tas)->state != TAPIOCA_SESSION_STATE_CLOSED, FALSE, error, TAPIOCA_ERROR, TAPIOCA_SESSION_ERROR_NOT_ESTABLISHED, "Session not established");
    g_set_error_val_if_fail (klass->add_local_candidate != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Not implemented");

    if (!klass->add_local_candidate (tas, 
                name,
                protocol,
                ip,
                port,
                preference,
                username,
                password,
                type,
                generation,
                error)) {
        if (*error == NULL)
            g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, NULL);
        return FALSE;
    }
    
    return TRUE;
}

/* workaround: remove the above functions when dbus works with child objects
   calling parent objects methods/signals */
gboolean 
tapioca_audio_session_initiate (TapiocaAudioSession *tas,
                                GError **error)
{
    return tapioca_session_initiate (TAPIOCA_SESSION (tas), error);
}

gboolean 
tapioca_audio_session_close (TapiocaAudioSession *tas,
                            GError **error)
{
    return tapioca_session_close (TAPIOCA_SESSION (tas), error);
}

gboolean 
tapioca_audio_session_get_session_type (TapiocaAudioSession *tas,
                                        guint *session_type,
                                        GError **error)
{
    return tapioca_session_get_session_type (TAPIOCA_SESSION (tas), session_type, error);
}

gboolean 
tapioca_audio_session_get_members (TapiocaAudioSession *tas,
                                  gchar ***ids,
                                  GError **error)
{
    return tapioca_session_get_members (TAPIOCA_SESSION (tas), ids, error);
}

gboolean 
tapioca_audio_session_add_members (TapiocaAudioSession *tas,
                                  const gchar **ids,
                                  GError **error)
{
    return tapioca_session_add_members (TAPIOCA_SESSION (tas), ids, error);
}

gboolean 
tapioca_audio_session_remove_members (TapiocaAudioSession *tas,
                                     const gchar **ids,
                                     GError **error)
{
    return tapioca_session_remove_members (TAPIOCA_SESSION (tas), ids, error);
}

gboolean
tapioca_audio_session_destroy (TapiocaAudioSession *tas,
                               GError **error)
{
    return tapioca_session_destroy (TAPIOCA_SESSION (tas), error);
}
