/*
 * Tapioca library
 * Copyright (C) 2005 INdT 
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapioca_connection.h"
#include "tapioca_connection_private.h"
#include "tapioca_connection_glue.h"

#include "tapioca_enums.h"
#include "tapioca_errors.h"
#include "tapioca_marshal.h"
#include "tapioca_utils.h"

#include <dbus/dbus-glib-bindings.h>

enum {
    INCOMING_SESSION,
    STATE_CHANGED,
    ERROR,
    FINISHED,
    LAST_SIGNAL
};

struct _TapiocaConnectionPrivate 
{
    GHashTable *sessions;
    TapiocaContactList *contact_list;
    TapiocaPresence *presence;
    GMutex *mutex;
    gboolean destroying;
    gboolean disconnecting;
    gboolean disposed;
};

G_DEFINE_TYPE(TapiocaConnection, tapioca_connection, G_TYPE_OBJECT)
    
static void tapioca_connection_dispose  (GObject *object);
static void tapioca_connection_finalize (GObject *object);

static void disconnect                  (TapiocaConnection *tc);
static void session_closed              (TapiocaSession *ts, 
                                         gpointer data);

static guint tapioca_connection_signals[LAST_SIGNAL] = { 0 };
static guint session_id = 0;

static void
tapioca_connection_class_init (TapiocaConnectionClass *klass)
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_connection_parent_class = g_type_class_peek_parent (klass);

    g_type_class_add_private (klass, sizeof (TapiocaConnectionPrivate));
   
    gobject_class->dispose = tapioca_connection_dispose;
    gobject_class->finalize = tapioca_connection_finalize;
    
    klass->connect = NULL;
    klass->disconnect = NULL;
    klass->create_session = NULL;
    
    /* Let's create our signals */
    tapioca_connection_signals[INCOMING_SESSION] = g_signal_new ("incoming-session",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            tapioca_marshal_VOID__STRING_UINT_STRING,
            G_TYPE_NONE,
            3,
            G_TYPE_STRING,
            G_TYPE_UINT,
            G_TYPE_STRING);
    tapioca_connection_signals[STATE_CHANGED] = g_signal_new ("state-changed",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            g_cclosure_marshal_VOID__UINT,
            G_TYPE_NONE,
            1,
            G_TYPE_UINT);
    tapioca_connection_signals[ERROR] = g_signal_new ("error",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            g_cclosure_marshal_VOID__UINT,
            G_TYPE_NONE,
            1,
            G_TYPE_UINT);
    tapioca_connection_signals[FINISHED] = g_signal_new ("finished",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            g_cclosure_marshal_VOID__VOID,
            G_TYPE_NONE,
            0,
            G_TYPE_NONE);

    dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (klass), 
            &dbus_glib_tapioca_connection_object_info);
}

static void
tapioca_connection_init (TapiocaConnection *tc)
{
    tc->state = TAPIOCA_CONNECTION_STATE_DISCONNECTED;
    tc->id = NULL;
    
    tc->priv = TAPIOCA_CONNECTION_GET_PRIVATE (tc);
    tc->priv->sessions = NULL;
    tc->priv->contact_list = NULL;
    tc->priv->presence = NULL;
    tc->priv->mutex = g_mutex_new ();
    tc->priv->destroying = FALSE;
    tc->priv->disconnecting = FALSE;
    tc->priv->disposed = FALSE;
}

static void
tapioca_connection_dispose (GObject *object)
{    
    TapiocaConnection *tc = TAPIOCA_CONNECTION (object);
    TapiocaConnectionClass *klass = TAPIOCA_CONNECTION_GET_CLASS (tc);

    if (tc->priv->disposed)
        return;

    tc->priv->disposed = TRUE;

    if (tc->priv->sessions)
        g_hash_table_destroy (tc->priv->sessions);

    // be sure we will delete the contact list object
    if (!tc->priv->presence && klass->get_presence)
        tc->priv->contact_list = klass->get_contact_list (tc);

    if (tc->priv->contact_list)
        g_object_unref (tc->priv->contact_list);

    // be sure we will delete the presence object
    if (!tc->priv->presence && klass->get_presence)
        tc->priv->presence = klass->get_presence (tc);
        
    if (tc->priv->presence)
        g_object_unref (tc->priv->presence);
    
    g_mutex_free (tc->priv->mutex);

    G_OBJECT_CLASS (tapioca_connection_parent_class)->dispose (object);
}

static void
tapioca_connection_finalize (GObject *object)
{
    g_signal_handlers_destroy (object);

    G_OBJECT_CLASS (tapioca_connection_parent_class)->finalize (object);
}

/**
 * Change the connection state.
 *
 * @param tc #TapiocaConnection instance
 * @param state #TapiocaConnectionState representing the new connection state
 */
void 
tapioca_connection_change_state (TapiocaConnection *tc,
                                 TapiocaConnectionState state)
{
    g_return_if_fail (tc != NULL);

    g_mutex_lock (tc->priv->mutex);
    
    if (tc->state == state) {
        g_mutex_unlock (tc->priv->mutex);
        return;
    }

    tc->state = state;

    if (!tc->priv->destroying) {
        g_mutex_unlock (tc->priv->mutex);
        g_signal_emit (tc, tapioca_connection_signals[STATE_CHANGED], 
                0, state);
        g_mutex_lock (tc->priv->mutex);
    }

    if ((tc->priv->destroying) && 
        (state == TAPIOCA_CONNECTION_STATE_DISCONNECTED)) {
        g_mutex_unlock (tc->priv->mutex);
        /* emit finished signal so ConnectionManager can delete this Connection */
        g_signal_emit (tc, tapioca_connection_signals[FINISHED], 0);
        return;
    }
    
    g_mutex_unlock (tc->priv->mutex);
}

/**
 * Emit a signal indicating an incoming session.
 *
 * @param tc #TapiocaConnection instance
 * @param from address of user who initiated the session
 * @param ts incoming #TapiocaSession instance
 */
void
tapioca_connection_signal_incoming_session (TapiocaConnection *tc,
                                            const gchar *from,
                                            TapiocaSession *ts)
{
    DBusGConnection *bus;
    GError *bus_error = NULL;
    gchar *object_path;

    g_mutex_lock (tc->priv->mutex);
    
    g_debug ("incoming session ...");
    
    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &bus_error);
    if (!bus) {
        g_warning ("unable to open connection to bus: %s", bus_error->message);
        g_mutex_unlock (tc->priv->mutex);
        return;
    }

    g_signal_connect (G_OBJECT (ts), "closed", (GCallback) session_closed, tc);

    if (!tc->priv->sessions)
        tc->priv->sessions = g_hash_table_new_full (g_direct_hash, g_direct_equal, 
                NULL, g_object_unref);
    g_hash_table_insert (tc->priv->sessions, ts, ts);

    object_path = g_strdup_printf ("/org/tapioca/Session/%d", ++session_id);    
    dbus_g_connection_register_g_object (bus,
            object_path,
            G_OBJECT (ts));
    
    if (!ts->members)
        ts->members = g_list_append (ts->members, (gpointer) g_strdup (from));

    g_signal_emit (tc, tapioca_connection_signals[INCOMING_SESSION], 
            0, from, ts->session_type, object_path);
    
    g_debug ("session created: session=%p object_path=%s", ts, object_path);

    g_free (object_path);
    
    g_mutex_unlock (tc->priv->mutex);
}

/**
 * Emit a signal indicating an error in the connection.
 * Note that when an error occurs the connection will be 
 * terminated automatically.
 *
 * @param tc #TapiocaConnection instance
 * @param error_code #TapiocaConnectionError error code
 */
void 
tapioca_connection_signal_error (TapiocaConnection *tc,
                                 guint error_code)
{
    g_return_if_fail (tc != NULL);

    g_signal_emit (tc, tapioca_connection_signals[ERROR], 0, error_code);
}

gboolean
tapioca_connection_connect (TapiocaConnection *tc, 
                            const gchar *id, 
                            const GHashTable *parameters, 
                            GError **error)
{
    TapiocaConnectionClass *klass = TAPIOCA_CONNECTION_GET_CLASS (tc);
    
    g_set_error_val_if_fail (klass->connect != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Function not implemented");
   
    g_debug ("connecting using id %s", id);

    g_mutex_lock (tc->priv->mutex);
    
    if (tc->state != TAPIOCA_CONNECTION_STATE_DISCONNECTED) {
        g_debug ("already connected or connecting!");
        g_mutex_unlock (tc->priv->mutex);
        return TRUE;
    }
    
    g_mutex_unlock (tc->priv->mutex);

    if (!klass->connect (tc, id, parameters, error)) {
        if (*error == NULL)
            g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "Unable to connect");
        return FALSE;
    }
   
    tc->id = g_strdup (id);

    return TRUE;
}

gboolean
tapioca_connection_disconnect (TapiocaConnection *tc, 
                               GError **error)
{
    g_mutex_lock (tc->priv->mutex);
    if (tc->state == TAPIOCA_CONNECTION_STATE_DISCONNECTED) {
        g_mutex_unlock (tc->priv->mutex);
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_CONNECTION_ERROR_NOT_CONNECTED, "Not connected");
        return FALSE;
    }
   
    if (tc->priv->disconnecting) {
        g_mutex_unlock (tc->priv->mutex);
        return TRUE;
    }
                
    g_debug ("disconnecting ...");
      
    tc->priv->disconnecting = TRUE;
    g_mutex_unlock (tc->priv->mutex);

    disconnect (tc);
    
    return TRUE;
}

gboolean
tapioca_connection_create_session (TapiocaConnection *tc, 
                                   guint session_type,
                                   gchar **object_path,
                                   GError **error)
{
    TapiocaConnectionClass *klass = TAPIOCA_CONNECTION_GET_CLASS (tc);
    GObject *ts;
    DBusGConnection *bus;
    GError *bus_error = NULL;

    g_set_error_val_if_fail (!(session_type != TAPIOCA_SESSION_TYPE_AUDIO && session_type != TAPIOCA_SESSION_TYPE_TEXT), FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_INVALID_PARAMETERS, "Invalid parameters");
    g_set_error_val_if_fail (klass->create_session != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Function not implemented");
    
    g_debug ("creating session of type %d ...", session_type);

    g_mutex_lock (tc->priv->mutex);
    if (tc->state != TAPIOCA_CONNECTION_STATE_CONNECTED) {
        g_mutex_unlock (tc->priv->mutex);
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_CONNECTION_ERROR_NOT_CONNECTED, "Not connected");
        return FALSE;
    }
    
    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &bus_error);
    if (!bus) {
        g_mutex_unlock (tc->priv->mutex);
        g_warning ("unable to open connection to bus: %s", bus_error->message);
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_ENOUGH_MEMORY, "Not enough memory");
        return FALSE;
    }
   
    ts = G_OBJECT (klass->create_session (tc, session_type, error));
    if (!ts) {
        g_mutex_unlock (tc->priv->mutex);
        if (*error == NULL)
            g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, NULL);
        return FALSE;    
    }
    g_signal_connect (ts, "closed", (GCallback) session_closed, tc);

    if (!tc->priv->sessions)
        tc->priv->sessions = g_hash_table_new_full (g_direct_hash, g_direct_equal, 
                NULL, g_object_unref);
    g_hash_table_insert (tc->priv->sessions, ts, ts);

    *object_path = g_strdup_printf ("/org/tapioca/Session/%d", ++session_id);
    dbus_g_connection_register_g_object (bus,
            *object_path,
            ts);
    
    g_debug ("session created: session=%p object_path=%s", ts, *object_path);
        
    g_mutex_unlock (tc->priv->mutex);

    return TRUE;
}

gboolean
tapioca_connection_get_presence (TapiocaConnection *tc, 
                                 gchar **object_path,
                                 GError **error)
{
    TapiocaConnectionClass *klass = TAPIOCA_CONNECTION_GET_CLASS (tc);
    DBusGConnection *bus;
    GError *bus_error = NULL;
    static guint presence_id = 0; 

    // TODO if the presence object is already created, just return the object_path
    
    if (!(klass->get_presence) ||
        !(tc->priv->presence = klass->get_presence (tc))) {
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_SUPPORTED, "Presence not support on this protocol");
        return FALSE;
    }
    
    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &bus_error);
    if (!bus) {
        g_warning ("unable to open connection to bus: %s", bus_error->message);
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_ENOUGH_MEMORY, "Not enough memory");
        return FALSE;
    }

    *object_path = g_strdup_printf ("/org/tapioca/Presence/%d", presence_id++);
    dbus_g_connection_register_g_object (bus,
                *object_path,
                G_OBJECT (tc->priv->presence));

    return TRUE;
}

gboolean
tapioca_connection_get_contact_list (TapiocaConnection *tc, 
                                     gchar **object_path,
                                     GError **error)
{
    TapiocaConnectionClass *klass = TAPIOCA_CONNECTION_GET_CLASS (tc);
    DBusGConnection *bus;
    GError *bus_error = NULL;
    static guint contact_list_id = 0; 

    // TODO if the presence object is already created, just return the object_path
    
    if (!(klass->get_contact_list) ||
        !(tc->priv->contact_list = klass->get_contact_list (tc))) {
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_SUPPORTED, "Contact List not support on this protocol");
        return FALSE;
    }
    
    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &bus_error);
    if (!bus) {
        g_warning ("unable to open connection to bus: %s", bus_error->message);
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_ENOUGH_MEMORY, "Not enough memory");
        return FALSE;
    }

    *object_path = g_strdup_printf ("/org/tapioca/ContactList/%d", contact_list_id++);
    dbus_g_connection_register_g_object (bus,
                *object_path,
                G_OBJECT (tc->priv->contact_list));

    return TRUE;
}

gboolean
tapioca_connection_destroy (TapiocaConnection *tc, 
                            GError **error)
{
    g_debug ("destroying connection ...");
    
    g_mutex_lock (tc->priv->mutex);

    if (tc->priv->destroying) {
        g_mutex_unlock (tc->priv->mutex);
        return TRUE;
    }
    
    tc->priv->destroying = TRUE;
    if ((!tc->priv->disconnecting) && 
        (tc->state != TAPIOCA_CONNECTION_STATE_DISCONNECTED)) {
        tc->priv->disconnecting = TRUE;
        g_mutex_unlock (tc->priv->mutex);
        disconnect (tc);
    }
    else {
        g_mutex_unlock (tc->priv->mutex);
        /* emit finished signal so ConnectionManager can delete this Connection */
        g_signal_emit (tc, tapioca_connection_signals[FINISHED], 0);
    }
        
    return TRUE;
}

static void
disconnect (TapiocaConnection *tc)
{
    TapiocaConnectionClass *klass = TAPIOCA_CONNECTION_GET_CLASS (tc);
    
    if (klass->disconnect)
        klass->disconnect (tc);
}

static void
session_closed (TapiocaSession *ts, 
                gpointer data)
{
    TapiocaConnection *tc = TAPIOCA_CONNECTION (data);

    g_debug ("removing session %p", ts);
    
    g_hash_table_remove (tc->priv->sessions, ts);
}
