/*
 * Tapioca daemon
 * Copyright (C) 2005 INdT 
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapioca_connection_manager.h"
#include "tapioca_connection_manager_glue.h"

#include <tapioca/core/tapioca_errors.h>
#include <tapioca/core/tapioca_utils.h>

struct _TapiocaConnectionManagerPrivate {
    GHashTable *connections;
    gboolean disposed;
};

static void tapioca_connection_manager_dispose    (GObject *object);

static void connection_finished                   (TapiocaConnection *conn, 
                                                   gpointer data);

G_DEFINE_TYPE(TapiocaConnectionManager, tapioca_connection_manager, G_TYPE_OBJECT)

static void
tapioca_connection_manager_class_init (TapiocaConnectionManagerClass *klass)
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_connection_manager_parent_class = g_type_class_peek_parent (klass);
    
    g_type_class_add_private (klass, sizeof (TapiocaConnectionManagerPrivate));

    gobject_class->dispose = tapioca_connection_manager_dispose;

    dbus_g_object_type_install_info (
            G_TYPE_FROM_CLASS (klass), 
            &dbus_glib_tapioca_connection_manager_object_info);
}

static void
tapioca_connection_manager_init (TapiocaConnectionManager *tcm)
{
    tcm->priv = TAPIOCA_CONNECTION_MANAGER_GET_PRIVATE (tcm);
    tcm->priv->connections = NULL;
    tcm->priv->disposed = FALSE;
}

static void 
tapioca_connection_manager_dispose (GObject *object)
{
    TapiocaConnectionManager *tcm = TAPIOCA_CONNECTION_MANAGER (object);

    if (tcm->priv->disposed) {
        /* If dispose did already run, return. */
        return;
    }
    
    /* Make sure dispose does not run twice. */
    tcm->priv->disposed = TRUE;

    G_OBJECT_CLASS (tapioca_connection_manager_parent_class)->dispose (object);
}

gboolean
tapioca_connection_manager_create_connection (TapiocaConnectionManager *tcm, 
                                              gchar **object_path,
                                              GError **error)
{
    TapiocaConnectionManagerClass *klass = TAPIOCA_CONNECTION_MANAGER_GET_CLASS (tcm);
    DBusGConnection *bus;
    GError *bus_error = NULL;
    TapiocaConnection *tc;
    static guint connection_id = 0; 
    
    g_set_error_val_if_fail (klass->create_connection != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Function not implemented");

    g_debug ("creating connection ...");

    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &bus_error);
    if (!bus) {
        g_warning ("unable to open connection to bus: %s", bus_error->message);
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_ENOUGH_MEMORY, "Not enough memory");  
        return FALSE;
    }
    
    if (!(tc = klass->create_connection (tcm, error))) {
        if (*error == NULL)
            g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "Unable to create connection");
        return FALSE;
    }
 
    g_signal_connect (tc, "finished", (GCallback) connection_finished, tcm);

    if (!tcm->priv->connections)
        tcm->priv->connections = g_hash_table_new_full (g_direct_hash, g_direct_equal, 
                NULL, g_object_unref);
    g_hash_table_insert (tcm->priv->connections, tc, tc);

    *object_path = g_strdup_printf ("/org/tapioca/Connection/%d", connection_id++);
    dbus_g_connection_register_g_object (bus,
            *object_path,
            G_OBJECT (tc));

    g_debug ("connection created: connection=%p object_path=%s", tc, *object_path);

    return TRUE;
}

static void
connection_finished (TapiocaConnection *tc, 
                     gpointer data)
{
    TapiocaConnectionManager *tcm = TAPIOCA_CONNECTION_MANAGER (data);

    g_debug ("removing connection %p", tc);
    
    g_hash_table_remove (tcm->priv->connections, tc);
}
