/*
 * Tapioca library
 * Copyright (C) 2005 INdT 
 * @author Luiz Augusto von Dentz <luiz.dentz@indt.org.br>
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapioca_presence.h"
#include "tapioca_presence_private.h"
#include "tapioca_presence_glue.h"

#include "tapioca_enums.h"
#include "tapioca_errors.h"
#include "tapioca_marshal.h"
#include "tapioca_utils.h"

#include <dbus/dbus-glib-bindings.h>

enum {
    SUBSCRIBED,
    UNSUBSCRIBED,
    STATUS_UPDATED,
    LAST_SIGNAL
};

struct _TapiocaPresencePrivate 
{
    GMutex *mutex;
    GHashTable *watch_list;
    gboolean subscribed;
    gboolean disposed;
};

G_DEFINE_TYPE(TapiocaPresence, tapioca_presence, G_TYPE_OBJECT)
    
static void  tapioca_presence_dispose  (GObject *object);
static void  tapioca_presence_finalize (GObject *object);

static guint tapioca_presence_signals[LAST_SIGNAL] = { 0 };

static void
tapioca_presence_class_init (TapiocaPresenceClass *klass)
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_presence_parent_class = g_type_class_peek_parent (klass);

    g_type_class_add_private (klass, sizeof (TapiocaPresencePrivate));
   
    gobject_class->dispose = tapioca_presence_dispose;
    gobject_class->finalize = tapioca_presence_finalize;
    
    klass->subscribe = NULL;
    klass->unsubscribe = NULL;
    klass->set_status = NULL;
    klass->get_supported_statuses = NULL;
    klass->request_status = NULL;
    klass->release_status = NULL;
        
    /* Let's create our signals */
    tapioca_presence_signals[SUBSCRIBED] = g_signal_new ("subscribed",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            g_cclosure_marshal_VOID__VOID,
            G_TYPE_NONE,
            0,
            G_TYPE_NONE);
    tapioca_presence_signals[UNSUBSCRIBED] = g_signal_new ("unsubscribed",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            g_cclosure_marshal_VOID__VOID,
            G_TYPE_NONE,
            0,
            G_TYPE_NONE);
    tapioca_presence_signals[STATUS_UPDATED] = g_signal_new ("status-updated",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            tapioca_marshal_VOID__STRING_UINT_STRING_BOOLEAN,
            G_TYPE_NONE,
            4,
            G_TYPE_STRING,
            G_TYPE_UINT,
            G_TYPE_STRING,
            G_TYPE_BOOLEAN);

    dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (klass), 
            &dbus_glib_tapioca_presence_object_info);
}

static void
tapioca_presence_init (TapiocaPresence *tp)
{
    tp->priv = TAPIOCA_PRESENCE_GET_PRIVATE (tp);
    tp->priv->mutex = g_mutex_new ();
    tp->priv->watch_list = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, g_free);
    tp->priv->subscribed = FALSE;
    tp->priv->disposed = FALSE;
}

static void
tapioca_presence_dispose (GObject *object)
{    
    TapiocaPresence *tp = TAPIOCA_PRESENCE (object);

    if (tp->priv->disposed)
        return;

    tp->priv->disposed = TRUE;
    
    g_mutex_free (tp->priv->mutex);
    g_hash_table_destroy (tp->priv->watch_list);

    G_OBJECT_CLASS (tapioca_presence_parent_class)->dispose (object);
}

static void
tapioca_presence_finalize (GObject *object)
{
    g_signal_handlers_destroy (object);

    G_OBJECT_CLASS (tapioca_presence_parent_class)->finalize (object);
}

gboolean
tapioca_presence_subscribe (TapiocaPresence *tp,
                            guint status,
                            const char *status_desc,
                            GError **error)
{
    TapiocaPresenceClass *klass = TAPIOCA_PRESENCE_GET_CLASS (tp);
    
    g_set_error_val_if_fail (klass->subscribe != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Function not implemented");

    g_debug ("subscribing ...");
    return klass->subscribe (tp, status, status_desc, error);
}

gboolean
tapioca_presence_unsubscribe (TapiocaPresence *tp,
                              GError **error)
{
    TapiocaPresenceClass *klass = TAPIOCA_PRESENCE_GET_CLASS (tp);
    
    g_set_error_val_if_fail (klass->unsubscribe != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Function not implemented");

    g_debug ("unsubscribing ...");
    return klass->unsubscribe (tp, error);
}

gboolean
tapioca_presence_get_supported_statuses (TapiocaPresence *tp,
                                         GArray **statuses,
                                         GError **error)
{
    TapiocaPresenceClass *klass = TAPIOCA_PRESENCE_GET_CLASS (tp);
    
    g_set_error_val_if_fail (klass->get_supported_statuses != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Function not implemented");

    g_debug ("getting supported statuses ...");
    klass->get_supported_statuses (tp, statuses);

    return TRUE;
}

gboolean
tapioca_presence_set_status (TapiocaPresence *tp,
                             guint status,
                             const char *status_desc,
                             GError **error)
{
    TapiocaPresenceClass *klass = TAPIOCA_PRESENCE_GET_CLASS (tp);
    
    g_set_error_val_if_fail (klass->set_status != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Function not implemented");

    g_debug ("setting new status...");
    return klass->set_status (tp, status, status_desc, error);
}

gboolean
tapioca_presence_request_status (TapiocaPresence *tp,
                                 gchar **list,
                                 GError **error)
{
    TapiocaPresenceClass *klass = TAPIOCA_PRESENCE_GET_CLASS (tp);
    guint i;
    gchar *contact;
    
    g_set_error_val_if_fail (klass->request_status != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Function not implemented");

    g_debug ("requesting status for:");
    g_mutex_lock (tp->priv->mutex);
    for (i = 0; list[i]; ++i) {
        if (!g_hash_table_lookup (tp->priv->watch_list, list[i])) {
            contact = g_strdup (list[i]);
            g_hash_table_insert (tp->priv->watch_list, contact, contact);
            g_debug ("contact '%s' inserted on watch list", list[i]);
        }
        else
            g_debug ("contact '%s' already on watch list", list[i]);
    }
    g_mutex_unlock (tp->priv->mutex);
    
    klass->request_status (tp, list);
    
    return TRUE;
}

gboolean
tapioca_presence_release_status (TapiocaPresence *tp,
                                 gchar **list,
                                 GError **error)
{
    TapiocaPresenceClass *klass = TAPIOCA_PRESENCE_GET_CLASS (tp);
    guint i;
    
    g_set_error_val_if_fail (klass->release_status != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Function not implemented");

    g_mutex_lock (tp->priv->mutex);
    g_debug ("released status for:");
    for (i = 0; list[i]; ++i) {
        g_hash_table_remove (tp->priv->watch_list, list[i]);
        g_debug ("contact '%s' removed from watch list", list[i]);
    }
    g_mutex_unlock (tp->priv->mutex);
    klass->release_status (tp, list);

    return TRUE;
}

gboolean
tapioca_presence_in_watch_list (TapiocaPresence *tp,
                                const gchar *contact)
{
    gboolean in = FALSE;
    
    g_debug ("checking if contact '%s' is on watch list", contact);
    
    g_mutex_lock (tp->priv->mutex);    
    if (g_hash_table_lookup (tp->priv->watch_list, contact)) {
        g_debug ("contact '%s' found on watch list", contact);
        in = TRUE;
    }
    else {
        g_debug ("contact '%s' not found on watch list", contact);
    }
    g_mutex_unlock (tp->priv->mutex);

    return in;
}

void
tapioca_presence_signal_subscribed (TapiocaPresence *tp)
{
    g_return_if_fail (tp != NULL);

    g_mutex_lock (tp->priv->mutex);
    tp->priv->subscribed = TRUE;
    g_mutex_unlock (tp->priv->mutex);
    g_signal_emit (tp, tapioca_presence_signals[SUBSCRIBED], 0);
    g_debug ("subscribed");
}

void
tapioca_presence_signal_status_updated (TapiocaPresence *tp,
                                        const gchar *contact,
                                        guint status,
                                        const gchar *status_desc,
                                        gboolean phone_capability)
{
    g_return_if_fail (tp != NULL);

    g_signal_emit (tp, tapioca_presence_signals[STATUS_UPDATED], 0, 
            contact, status, status_desc, phone_capability);
    g_debug ("status updated for contact %s: status=%d status_desc=%s phone_capability=%s", 
            contact, status, status_desc, (phone_capability ? "TRUE" : "FALSE"));
}

void
tapioca_presence_signal_unsubscribed (TapiocaPresence *tp)
{
    g_return_if_fail (tp != NULL);

    g_mutex_lock (tp->priv->mutex);
    if (tp->priv->subscribed) {
        tp->priv->subscribed = FALSE;
        g_mutex_unlock (tp->priv->mutex);
        g_signal_emit (tp, tapioca_presence_signals[UNSUBSCRIBED], 0);
        g_debug ("unsubscribed");
    }
    else {
        g_debug ("not subscribed");
        g_mutex_unlock (tp->priv->mutex);
    }
}
