/*
 * Tapioca library
 * Copyright (C) 2005 INdT 
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapioca_session.h"
#include "tapioca_session_private.h"
/* workaround: add this when dbus works with child objects 
   calling parent objects methods/signals */
// #include "tapioca_session_glue.h"

#include "tapioca_enums.h"
#include "tapioca_errors.h"
#include "tapioca_marshal.h"
#include "tapioca_utils.h"

#include <glib-object.h>
#include <dbus/dbus-glib-bindings.h>

enum {
    MEMBER_ADDED,
    MEMBER_REMOVED,
    ADD_MEMBER_ERROR,
    STATE_CHANGED,
    ERROR,
    CLOSED,
    LAST_SIGNAL
};

struct _TapiocaSessionPrivate 
{
    GMutex *mutex;
    gboolean destroying;
    gboolean closing;
    gboolean disposed;
};
    
G_DEFINE_TYPE(TapiocaSession, tapioca_session, G_TYPE_OBJECT)
    
static void tapioca_session_dispose  (GObject *object);
static void tapioca_session_finalize (GObject *object);

static void close_session            (TapiocaSession *ts);

static guint tapioca_session_signals[LAST_SIGNAL] = { 0 };

static void
tapioca_session_class_init (TapiocaSessionClass *klass)
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_session_parent_class = g_type_class_peek_parent (klass);

    g_type_class_add_private (klass, sizeof (TapiocaSessionPrivate));
   
    gobject_class->dispose = tapioca_session_dispose;
    gobject_class->finalize = tapioca_session_finalize;
    
    klass->initiate = NULL;
    klass->close = NULL;
    klass->add_members = NULL;
    klass->remove_members = NULL;
    
    /* Let's create our signals */
    tapioca_session_signals[MEMBER_ADDED] = g_signal_new ("member-added",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            g_cclosure_marshal_VOID__STRING,
            G_TYPE_NONE,
            1,
            G_TYPE_STRING);
    tapioca_session_signals[MEMBER_REMOVED] = g_signal_new ("member-removed",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            g_cclosure_marshal_VOID__STRING,
            G_TYPE_NONE,
            1,
            G_TYPE_STRING);
    tapioca_session_signals[ADD_MEMBER_ERROR] = g_signal_new ("add-member-error",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            tapioca_marshal_VOID__UINT_STRING,
            G_TYPE_NONE,
            2,
            G_TYPE_UINT,
            G_TYPE_STRING);
    tapioca_session_signals[STATE_CHANGED] = g_signal_new ("state-changed",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            g_cclosure_marshal_VOID__UINT,
            G_TYPE_NONE,
            1,
            G_TYPE_UINT);
    tapioca_session_signals[ERROR] = g_signal_new ("error",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            g_cclosure_marshal_VOID__UINT,
            G_TYPE_NONE,
            1,
            G_TYPE_UINT);
     tapioca_session_signals[CLOSED] = g_signal_new ("closed",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            g_cclosure_marshal_VOID__VOID,
            G_TYPE_NONE,
            0,
            G_TYPE_NONE);
   
    /* workaround: add this when dbus works with child objects 
       calling parent objects methods/signals */
    // dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (klass), 
    //         &dbus_glib_tapioca_session_object_info);
}

static void
tapioca_session_init (TapiocaSession *ts)
{
    ts->session_type = TAPIOCA_SESSION_TYPE_UNKNOWN;
    ts->state = TAPIOCA_SESSION_STATE_CLOSED;
    ts->members = NULL;

    ts->priv = TAPIOCA_SESSION_GET_PRIVATE (ts);
    ts->priv->mutex = g_mutex_new ();
    ts->priv->destroying = FALSE;
    ts->priv->closing = FALSE;
    ts->priv->disposed = FALSE;
}

static void
tapioca_session_dispose (GObject *object)
{    
    TapiocaSession *ts = TAPIOCA_SESSION (object);

    if (ts->priv->disposed)
        return;

    ts->priv->disposed = TRUE;

    G_OBJECT_CLASS (tapioca_session_parent_class)->dispose (object);
}

static void
tapioca_session_finalize (GObject *object)
{
    g_signal_handlers_destroy (object);

    G_OBJECT_CLASS (tapioca_session_parent_class)->finalize (object);
}

void 
tapioca_session_change_state (TapiocaSession *ts,
                              guint state)
{
    g_return_if_fail (ts != NULL);

    g_mutex_lock (ts->priv->mutex);
    
    if (ts->state == state) {
        g_mutex_unlock (ts->priv->mutex);
        return;
    }
    
    ts->state = state;    

    if (!ts->priv->destroying) {
        g_mutex_unlock (ts->priv->mutex);
        g_signal_emit (ts, tapioca_session_signals[STATE_CHANGED], 
                0, state);
        g_mutex_lock (ts->priv->mutex);
    }

    if ((ts->priv->destroying) && 
        (state == TAPIOCA_SESSION_STATE_CLOSED)) {
        g_mutex_unlock (ts->priv->mutex);
        /* emit finished signal so Connection can delete this Session */
        g_signal_emit (ts, tapioca_session_signals[CLOSED], 0);
        return;
    }
    
    g_mutex_unlock (ts->priv->mutex);
}

/**
 * Emit a signal indicating that a member has been added to session.
 *
 * @param ts #TapiocaSession instance
 * @param id id representing the member added
 */
void 
tapioca_session_signal_member_added (TapiocaSession *ts,
                                     const gchar *id)
{
    g_debug ("member %s added to session", id);

    g_mutex_lock (ts->priv->mutex);
    ts->members = g_list_append (ts->members, (gpointer) g_strdup (id));
    g_mutex_unlock (ts->priv->mutex);

    g_signal_emit (ts, tapioca_session_signals[MEMBER_ADDED], 0, id);
}

/**
 * Emit a signal indicating that a member has been removed from session.
 * Note that if this is the last member in the session, the session 
 * will be closed.
 *
 * @param ts #TapiocaSession instance
 * @param id id representing the member removed
 */
void 
tapioca_session_signal_member_removed (TapiocaSession *ts,
                                       const gchar *id)
{
    g_mutex_lock (ts->priv->mutex);
            
    g_debug ("trying to remove member %s from session", id);
        
    /* do not remove the last member of the session */
    if (g_list_length (ts->members) > 0) {
        GList *member;
 
        member = g_list_find_custom (ts->members, id, (GCompareFunc) g_strcasecmp);
        if (!member) {
            g_debug ("trying to remove member that is not in the session");
            g_mutex_unlock (ts->priv->mutex);
            return;
        }
        
        g_debug ("member %s removed from session", id);
        g_free (member->data);
        ts->members = g_list_remove_link (ts->members, member);
        
        g_mutex_unlock (ts->priv->mutex);
        
        g_signal_emit (ts, tapioca_session_signals[MEMBER_REMOVED], 0, id);
    }
        
    g_mutex_unlock (ts->priv->mutex);
}

/**
 * Emit a signal indicating an error adding a member to session.
 *
 * @param ts #TapiocaSession instance
 * @param error_code error code
 * @param id id representing the member that could not be added
 */
void 
tapioca_session_signal_add_member_error (TapiocaSession *ts,
                                         guint error_code,
                                         const gchar *id)
{
    g_signal_emit (ts, tapioca_session_signals[ADD_MEMBER_ERROR], 
            0, error_code, id);
}

/**
 * Emit a signal indicating an error in the session.
 * Note that when an error occurs the session will be 
 * closed automatically.
 *
 * @param ts #TapiocaSession instance
 * @param error_code #TapiocaSessionError error code
 */
void 
tapioca_session_signal_error (TapiocaSession *ts,
                              guint error_code)
{
    g_return_if_fail (ts != NULL);

    g_signal_emit (ts, tapioca_session_signals[ERROR], 0, error_code);
}

gboolean
tapioca_session_initiate (TapiocaSession *ts, 
                          GError **error)
{
    TapiocaSessionClass *klass = TAPIOCA_SESSION_GET_CLASS (ts);
    
    g_set_error_val_if_fail (klass->initiate != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Not implemented");
    g_set_error_val_if_fail (ts->members != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "No member added to session");

    g_mutex_lock (ts->priv->mutex);
    
    if (ts->state != TAPIOCA_SESSION_STATE_CLOSED) {
        g_mutex_unlock (ts->priv->mutex);
        g_debug ("already established or initiating!");
        return TRUE;
    }
    
    g_mutex_unlock (ts->priv->mutex);

    if (!klass->initiate (ts, error)) {
        if (*error == NULL)
            g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, "Unable to initiate");
        return FALSE;
    }
        
    return TRUE;
}

gboolean
tapioca_session_close (TapiocaSession *ts, 
                       GError **error)
{
    g_mutex_lock (ts->priv->mutex);

    if (ts->state == TAPIOCA_SESSION_STATE_CLOSED) {
        g_mutex_unlock (ts->priv->mutex);
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_SESSION_ERROR_NOT_ESTABLISHED, "Session not established");
        return FALSE;
    }

    if (ts->priv->closing) {
        g_mutex_unlock (ts->priv->mutex);
        return TRUE;
    }
    
    g_debug ("closing session ...");
   
    ts->priv->closing = TRUE;
    g_mutex_unlock (ts->priv->mutex);
    close_session (ts);

    return TRUE;
}

gboolean
tapioca_session_get_session_type (TapiocaSession *ts,
                                  guint *session_type,
                                  GError **error)
{
    *session_type = ts->session_type;
    return TRUE;
}

gboolean
tapioca_session_get_members (TapiocaSession *ts, 
                             gchar ***ids,
                             GError **error)
{
    GList *members_walk;
    guint length = 0;
    guint i = 0;
    
    g_mutex_lock (ts->priv->mutex);
        
    if (ts->members)
        length = g_list_length (ts->members);
    *ids = g_new (gchar *, length + 1);
    for (members_walk = ts->members; members_walk != NULL; 
            members_walk = g_list_next (members_walk)) 
        (*ids)[i++] = g_strdup (((const gchar *) members_walk->data));
    (*ids)[i] = NULL;
    
    g_mutex_unlock (ts->priv->mutex);
    
    return TRUE;
}

gboolean
tapioca_session_add_members (TapiocaSession *ts, 
                             const gchar **ids,
                             GError **error)
{    
    TapiocaSessionClass *klass = TAPIOCA_SESSION_GET_CLASS (ts);
    
    g_set_error_val_if_fail (klass->add_members != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Not implemented");

    if (!klass->add_members (ts, ids, error)) {
        if (*error == NULL)
            g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, NULL);
        return FALSE;
    }

    return TRUE;
}

gboolean
tapioca_session_remove_members (TapiocaSession *ts, 
                                const gchar **ids,
                                GError **error)
{
    TapiocaSessionClass *klass = TAPIOCA_SESSION_GET_CLASS (ts);
    
    g_set_error_val_if_fail (klass->remove_members != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Not implemented");
    
    return klass->remove_members (ts, ids, error);
}

gboolean
tapioca_session_destroy (TapiocaSession *ts, 
                         GError **error)
{
    g_debug ("destroying session ...");
    
    g_mutex_lock (ts->priv->mutex);

    if (ts->priv->destroying) {
        g_mutex_unlock (ts->priv->mutex);
        return TRUE;
    }
    
    ts->priv->destroying = TRUE;
    if ((!ts->priv->closing) && 
        (ts->state != TAPIOCA_SESSION_STATE_CLOSED)) {
        ts->priv->closing = TRUE;
        g_mutex_unlock (ts->priv->mutex);
        close_session (ts);
    }
    else {
        g_mutex_unlock (ts->priv->mutex);
        /* emit finished signal so Connection can delete this Session */
        g_signal_emit (ts, tapioca_session_signals[CLOSED], 0);
    }
        
    return TRUE;
}

static void
close_session (TapiocaSession *ts)
{
    TapiocaSessionClass *klass = TAPIOCA_SESSION_GET_CLASS (ts);

    if (klass->close)
        klass->close (ts);
}
