/*
 * Tapioca library
 * Copyright (C) 2005 INdT 
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tapioca_text_session.h"
#include "tapioca_text_session_private.h"
#include "tapioca_text_session_glue.h"

/* workaround: remove this when dbus works with child objects
   calling parent objects methods/signals */
#include "tapioca_session_private.h"

#include "tapioca_enums.h"
#include "tapioca_errors.h"
#include "tapioca_marshal.h"
#include "tapioca_utils.h"

#include <glib-object.h>
#include <dbus/dbus-glib-bindings.h>

#include <time.h>

enum {
    INCOMING,
    DELIVERY_ERROR,
    LAST_SIGNAL
};

struct _TapiocaTextSessionPrivate 
{
    GMutex *mutex;
    GList *queued_messages;
    gboolean enable_message_queue;
    gboolean disposed;
};

typedef struct _TapiocaTextMessage TapiocaTextMessage;
struct _TapiocaTextMessage {
    gchar *member;
    gchar *subject;
    gchar *body;
    guint type;
    guint timestamp;
    gboolean is_error;
    guint error_code;
};

G_DEFINE_TYPE(TapiocaTextSession, tapioca_text_session, TAPIOCA_TYPE_SESSION)
    
static void tapioca_text_session_dispose  (GObject *object);
static void tapioca_text_session_finalize (GObject *object);

static void destroy_message               (TapiocaTextMessage *message);

static guint tapioca_text_session_signals[LAST_SIGNAL] = { 0 };

static void
tapioca_text_session_class_init (TapiocaTextSessionClass *klass)
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_text_session_parent_class = g_type_class_peek_parent (klass);

    g_type_class_add_private (klass, sizeof (TapiocaTextSessionPrivate));
   
    gobject_class->dispose = tapioca_text_session_dispose;
    gobject_class->finalize = tapioca_text_session_finalize;
    
    klass->send = NULL;
    
    /* Let's create our signals */
    tapioca_text_session_signals[INCOMING] = g_signal_new ("incoming",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            tapioca_marshal_VOID__STRING_UINT_STRING_STRING_UINT,
            G_TYPE_NONE,
            5,
            G_TYPE_STRING,
            G_TYPE_UINT,
            G_TYPE_STRING,
            G_TYPE_STRING,
            G_TYPE_UINT);
    tapioca_text_session_signals[DELIVERY_ERROR] = g_signal_new ("delivery-error",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
            0,
            NULL,
            NULL,
            tapioca_marshal_VOID__STRING_UINT_STRING_STRING_UINT_UINT,
            G_TYPE_NONE,
            6,
            G_TYPE_STRING, 
            G_TYPE_UINT,
            G_TYPE_STRING, 
            G_TYPE_STRING, 
            G_TYPE_UINT,
            G_TYPE_UINT);
    
    dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (klass), 
            &dbus_glib_tapioca_text_session_object_info);
}

static void
tapioca_text_session_init (TapiocaTextSession *tts)
{
    TapiocaSession *ts = TAPIOCA_SESSION (tts);
    ts->session_type = TAPIOCA_SESSION_TYPE_TEXT;

    tts->priv = TAPIOCA_TEXT_SESSION_GET_PRIVATE (tts);
    tts->priv->mutex = g_mutex_new ();
    tts->priv->queued_messages = NULL;
    tts->priv->enable_message_queue = TRUE;
    tts->priv->disposed = FALSE;
}

static void
tapioca_text_session_dispose (GObject *object)
{    
    TapiocaTextSession *tts = TAPIOCA_TEXT_SESSION (object);

    if (tts->priv->disposed)
        return;

    tts->priv->disposed = TRUE;
    
    g_mutex_free (tts->priv->mutex);
    
    G_OBJECT_CLASS (tapioca_text_session_parent_class)->dispose (object);
}

static void
tapioca_text_session_finalize (GObject *object)
{
    g_signal_handlers_destroy (object);

    G_OBJECT_CLASS (tapioca_text_session_parent_class)->finalize (object);
}

void
tapioca_text_session_signal_incoming (TapiocaTextSession *tts,
                                      const gchar *member,
                                      guint message_type,
                                      const gchar *subject,
                                      const gchar *body)
{
    time_t timestamp;
    TapiocaTextMessage *message;
    
    g_return_if_fail (tts != NULL);
    g_return_if_fail (member != NULL);
    g_return_if_fail (body != NULL);
    
    g_mutex_lock (tts->priv->mutex);
    
    timestamp = time (NULL);
    
    if (!tts->priv->enable_message_queue) {
        g_signal_emit (tts, tapioca_text_session_signals[INCOMING], 
                0, member, message_type, subject, body, timestamp);
        g_mutex_unlock (tts->priv->mutex);
        return;
    }

    // add message to queue
    message = g_new (TapiocaTextMessage, 1);
    message->member = g_strdup (member);
    message->subject = g_strdup (subject);
    message->body = g_strdup (body);
    message->type = message_type;
    message->timestamp = timestamp;
    message->is_error = FALSE;
    message->error_code = 0;
    tts->priv->queued_messages = g_list_append (tts->priv->queued_messages, message);
    
    g_mutex_unlock (tts->priv->mutex);
}

void
tapioca_text_session_signal_delivery_error (TapiocaTextSession *tts,
                                            const gchar *member,
                                            guint message_type,
                                            const gchar *subject,
                                            const gchar *body,
                                            guint error_code)
{
    time_t timestamp;
    TapiocaTextMessage *message;

    g_return_if_fail (tts != NULL);
    g_return_if_fail (member != NULL);
    g_return_if_fail (body != NULL);

    g_mutex_lock (tts->priv->mutex);
    
    timestamp = time (NULL);
    
    if (!tts->priv->enable_message_queue) {
        g_signal_emit (tts, tapioca_text_session_signals[DELIVERY_ERROR], 
                0, member, message_type, subject, body, timestamp, error_code);
        g_mutex_unlock (tts->priv->mutex);
        return;
    }

    // add message to queue
    message = g_new (TapiocaTextMessage, 1);
    message->member = g_strdup (member);
    message->subject = g_strdup (subject);
    message->body = g_strdup (body);
    message->type = message_type;
    message->timestamp = timestamp;
    message->is_error = TRUE;
    message->error_code = error_code;
    tts->priv->queued_messages = g_list_append (tts->priv->queued_messages, message);
    
    g_mutex_unlock (tts->priv->mutex);
}

gboolean
tapioca_text_session_send (TapiocaTextSession *tts, 
                           guint message_type,
                           const gchar *subject,
                           const gchar *body,
                           GError **error)
{
    TapiocaTextSessionClass *klass = TAPIOCA_TEXT_SESSION_GET_CLASS (tts);
    
    g_set_error_val_if_fail (TAPIOCA_SESSION (tts)->state == TAPIOCA_SESSION_STATE_ESTABLISHED, FALSE, error, TAPIOCA_ERROR, TAPIOCA_SESSION_ERROR_NOT_ESTABLISHED, "Session not established");
    g_set_error_val_if_fail (klass->send != NULL, FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_IMPLEMENTED, "Not implemented");

    if (message_type != TAPIOCA_TEXT_SESSION_MESSAGE_TYPE_PLAIN_TEXT &&
        message_type != TAPIOCA_TEXT_SESSION_MESSAGE_TYPE_IS_TYPING) {
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_INVALID_PARAMETERS, "Invalid parameters");
        return FALSE;
    }
            
    if (!klass->send (tts, message_type, subject, body, error)) {
        if (*error == NULL)
            g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_GENERIC, NULL);
        return FALSE;
    }

    return TRUE;
}

gboolean
tapioca_text_session_set_enable_message_queue (TapiocaTextSession *tts,
                                               gboolean enable,
                                               GError *error)
{
    GList *qm;
    TapiocaTextMessage *message;

    g_mutex_lock (tts->priv->mutex);
    if (!enable) {
        for (qm = tts->priv->queued_messages; qm; qm = g_list_next (qm)) {
            message = (TapiocaTextMessage *) qm->data;       
            if (!message->is_error)
                g_signal_emit (tts, tapioca_text_session_signals[INCOMING], 
                    0, message->member, message->type, message->subject, message->body, message->timestamp);
            else
                g_signal_emit (tts, tapioca_text_session_signals[DELIVERY_ERROR], 
                    0, message->member, message->type, message->subject, message->body, message->timestamp, message->error_code);
            destroy_message (message);
        }
        if (tts->priv->queued_messages) {
            g_list_free (tts->priv->queued_messages);
            tts->priv->queued_messages = NULL;
        }
    }
       
    tts->priv->enable_message_queue = enable;
    g_mutex_unlock (tts->priv->mutex);
        
    return TRUE;
}

gboolean
tapioca_text_session_clear_message_queue (TapiocaTextSession *tts,
                                          GError *error)
{
    GList *qm;
    TapiocaTextMessage *message; 
    
    g_mutex_lock (tts->priv->mutex);
    if (!tts->priv->queued_messages) {
        g_mutex_unlock (tts->priv->mutex);
        return TRUE;
    }

    for (qm = tts->priv->queued_messages; qm; qm = g_list_next (qm)) {
        message = (TapiocaTextMessage *) qm->data;        
        destroy_message (message);
    }
    g_list_free (tts->priv->queued_messages);
    tts->priv->queued_messages = NULL;

    g_mutex_unlock (tts->priv->mutex);
        
    return TRUE;
}

/* workaround: remove the above functions when dbus works with child objects
   calling parent objects methods/signals */
gboolean 
tapioca_text_session_initiate (TapiocaTextSession *tts,
                               GError **error)
{
    return tapioca_session_initiate (TAPIOCA_SESSION (tts), error);
}

gboolean 
tapioca_text_session_close (TapiocaTextSession *tts,
                            GError **error)
{
    return tapioca_session_close (TAPIOCA_SESSION (tts), error);
}

gboolean 
tapioca_text_session_get_session_type (TapiocaTextSession *tts,
                                       guint *session_type,
                                       GError **error)
{
    return tapioca_session_get_session_type (TAPIOCA_SESSION (tts), session_type, error);
}

gboolean 
tapioca_text_session_get_members (TapiocaTextSession *tts,
                                  gchar ***ids,
                                  GError **error)
{
    return tapioca_session_get_members (TAPIOCA_SESSION (tts), ids, error);
}

gboolean 
tapioca_text_session_add_members (TapiocaTextSession *tts,
                                  const gchar **ids,
                                  GError **error)
{
    return tapioca_session_add_members (TAPIOCA_SESSION (tts), ids, error);
}

gboolean 
tapioca_text_session_remove_members (TapiocaTextSession *tts,
                                     const gchar **ids,
                                     GError **error)
{
    return tapioca_session_remove_members (TAPIOCA_SESSION (tts), ids, error);
}

gboolean
tapioca_text_session_destroy (TapiocaTextSession *tts,
                              GError **error)
{
    return tapioca_session_destroy (TAPIOCA_SESSION (tts), error);
}

static void
destroy_message (TapiocaTextMessage *message)
{
    g_free (message->member);
    g_free (message->subject);
    g_free (message->body);
    g_free (message);
}
